/* fold.cc: Definitions for member functions do constant folding */

/* This is done after resolve2, so types are available, but not checked
   yet */

#include "AST.h"
#include "compiler.h"
#include "template.h"
#include "decls.h"
#include "utils.h"

extern bool isCastNode(const TreeNode *t);

#define intMask (Common::Kind)(Common::CharKind | Common::ByteKind | \
			       Common::ShortKind |  Common::IntKind | \
			       Common::LongKind)

#define arithMask (Common::Kind)(intMask | Common::FloatKind | Common::DoubleKind)

#define primitiveMask (Common::Kind)(arithMask | Common::BoolKind)

static bool foldedFinalField;

void foldConstants()
{
  // Could sort things to make this faster, of course. But probably
  // not worth the effort.

  // Fold all final fields first, once that is done one pass suffices
  do
    {
      foldedFinalField = false;
      foreach (f, llist<CompileUnitNode *>, *allFiles)
	(*f)->foldFields();
      templateEnv.foldFields();
    }
  while (foldedFinalField);

  foreach (f, llist<CompileUnitNode *>, *allFiles)
    (*f)->foldConstants();
  templateEnv.foldConstants();

  assert(!foldedFinalField);
}

void foldInstantiations()
{
  llist< TreeNode * > * const instances = templateEnv.allInstances();
  
  do
    {
      foldedFinalField = false;
      foreach (f, llist<TreeNode *>, *instances)
	(*f)->foldFields();
    }
  while (foldedFinalField);

  foreach (f, llist<TreeNode *>, *instances)
    (*f)->foldConstants();

  assert(!foldedFinalField);
  free_all( instances );
}

// Fold fields only.
void TreeNode::foldFields()
{
  foriter (p, allChildren(), ChildIter)
    (*p)->foldFields();
}

// Fold fields, stop recursion in all other children of classes/interfaces
void FieldDeclNode::foldFields()
{
  foldConstants();
}

void StaticInitNode::foldFields()
{
}

void MethodNode::foldFields()
{
}

void ConstructorDeclNode::foldFields()
{
}

TreeNode *TreeNode::foldConstants()
{
  int fmask = foldMask();
  const int numChildren = arity();

  for (int sweep = 0; sweep < numChildren; ++sweep)
    {
      child(sweep, child(sweep)->foldConstants());

      fmask &= child(sweep)->constantType();
    }

  if (fmask)
    {
#if 0
      cout << "Folding\n";
      print(cout);
      TreeNode *folded = _fold();
      cout << "giving\n";
      folded->print(cout);

      return folded;
#else
      return _fold();
#endif
    }
  else return this;
}


TreeNode *TemplateDeclNode::foldConstants()
{
  return this;
}


TreeNode *CompileUnitNode::foldConstants()
{
  foriter (type, types()->allChildren(), ChildIter)
    (*type)->foldConstants();

  return this;
}


void TemplateDeclNode::foldFields()
{
}


TreeNode *FieldDeclNode::foldConstants()
{
  dtype()->foldConstants();

  TreeNode *oldInitExpr = initExpr();
  TreeNode *newInitExpr = oldInitExpr->foldConstants();

  initExpr(newInitExpr);
  if (oldInitExpr != newInitExpr && (flags() & Final))
    foldedFinalField = true;

  return this;
}

TreeNode *TreeNode::_fold()
{
  return this;
}

Common::Kind TreeNode::constantType()
{
  return (Common::Kind)0;
}

Common::Kind PrimitiveLitNode::constantType()
{
  return literal().kind();
}

Common::Kind TypeNode::constantType()
{
  // Hack: avoid typenodes confusing constant folding (cf cast)
  return primitiveMask;
}

Common::Kind TreeNode::foldMask()
{
  return (Common::Kind)0;
}

/* Arithmetic constant folding */

Common::Kind ComplementNode::foldMask()
{
  return intMask;
}

TreeNode *ComplementNode::_fold()
{
  return new PrimitiveLitNode(~opnd0()->literal(), position());
}

Common::Kind NotNode::foldMask()
{
  return Common::BoolKind;
}

TreeNode *NotNode::_fold()
{
  return new PrimitiveLitNode(!opnd0()->literal(), position());
}

Common::Kind CastNode::foldMask()
{
  return primitiveMask;
}

TreeNode *CastNode::_fold()
{
  TypeNode *dt = dtype();

  if (dt->isCastableFrom(opnd0()->type()))
    return new PrimitiveLitNode(opnd0()->literal().cast(dt->kind()), position());
  else
    return this;
}

TreeNode *IfExprNode::foldConstants()
{
  const int numChildren = arity();
  for (int sweep = 0; sweep < numChildren; ++sweep)
    child(sweep, child(sweep)->foldConstants());

  TreeNode *cond = condition();
  if (cond->constantType() == Common::BoolKind)
    if (cond->literal().boolValue()) return thenOpnd();
    else return elseOpnd();
  else return this;
}

TreeNode *IfStmtNode::foldConstants()
{
  const int numChildren = arity();
  for (int sweep = 0; sweep < numChildren; ++sweep)
    child(sweep, child(sweep)->foldConstants());

  TreeNode *cond = condition();
  if (cond->constantType() == Common::BoolKind) {
    TreeNode *result = cond->literal().boolValue() ? thenPart() : elsePart();
    return result->absent() ? new EmptyStmtNode(position()) : result;
  }
  else return this;
}

Common::Kind UnaryPlusNode::foldMask()
{
  return arithMask;
}

TreeNode *UnaryPlusNode::_fold()
{
  return opnd0();
}

Common::Kind UnaryMinusNode::foldMask()
{
  return arithMask;
}

TreeNode *UnaryMinusNode::_fold()
{
  return new PrimitiveLitNode(-opnd0()->literal(), position());
}

Common::Kind MultNode::foldMask()
{
  return arithMask;
}

TreeNode *MultNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal() * opnd1()->literal(), position());
}

Common::Kind DivNode::foldMask()
{
  return arithMask;
}

static bool notZero(Literal l)
{
  switch (l.kind())
    {
    default: return true;
    case IntegerKind:
      return l.intValue() != 0;
    case Common::FloatKind:
    case Common::DoubleKind:
      return l.doubleValue() != 0.0;
    }
}

TreeNode *DivNode::_fold()
{
  if (notZero(opnd1()->literal()))
    return new PrimitiveLitNode(opnd0()->literal() / opnd1()->literal(), position());
  else
    return this;
}

Common::Kind RemNode::foldMask()
{
  return arithMask;
}

TreeNode *RemNode::_fold()
{
  if (notZero(opnd1()->literal()))
    return new PrimitiveLitNode(opnd0()->literal() % opnd1()->literal(), position());
  else
    return this;
}

Common::Kind PlusNode::foldMask()
{
  return arithMask;
}

TreeNode *PlusNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal() + opnd1()->literal(), position());
}

Common::Kind MinusNode::foldMask()
{
  return arithMask;
}

TreeNode *MinusNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal() - opnd1()->literal(), position());
}

Common::Kind LeftShiftLogNode::foldMask()
{
  return intMask;
}

TreeNode *LeftShiftLogNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal().lsl(opnd1()->literal()), position());
}

Common::Kind RightShiftLogNode::foldMask()
{
  return intMask;
}

TreeNode *RightShiftLogNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal().rsl(opnd1()->literal()), position());
}

Common::Kind RightShiftArithNode::foldMask()
{
  return intMask;
}

TreeNode *RightShiftArithNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal().rsa(opnd1()->literal()), position());
}

Common::Kind LTNode::foldMask()
{
  return arithMask;
}

TreeNode *LTNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal().lt(opnd1()->literal()), position());
}

Common::Kind LENode::foldMask()
{
  return arithMask;
}

TreeNode *LENode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal().le(opnd1()->literal()), position());
}

Common::Kind GTNode::foldMask()
{
  return arithMask;
}

TreeNode *GTNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal().gt(opnd1()->literal()), position());
}

Common::Kind GENode::foldMask()
{
  return arithMask;
}

TreeNode *GENode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal().ge(opnd1()->literal()), position());
}

Common::Kind EQNode::foldMask()
{
  return primitiveMask;
}

TreeNode *EQNode::_fold()
{
  if ((opnd0()->constantType() == Common::BoolKind) == 
      (opnd1()->constantType() == Common::BoolKind))
    return new PrimitiveLitNode(opnd0()->literal().eq(opnd1()->literal()), position());
  else return this;

}

Common::Kind NENode::foldMask()
{
  return primitiveMask;
}

TreeNode *NENode::_fold()
{
  if ((opnd0()->constantType() == Common::BoolKind) == 
      (opnd1()->constantType() == Common::BoolKind))
    return new PrimitiveLitNode(opnd0()->literal().ne(opnd1()->literal()), position());
  else return this;

}

Common::Kind BitAndNode::foldMask()
{
  return (Common::Kind)(Common::BoolKind | intMask);
}

TreeNode *BitAndNode::_fold()
{
  if ((opnd0()->constantType() == Common::BoolKind) == 
      (opnd1()->constantType() == Common::BoolKind))
    return new PrimitiveLitNode(opnd0()->literal() & opnd1()->literal(), position());
  else return this;

}

Common::Kind BitOrNode::foldMask()
{
  return (Common::Kind)(Common::BoolKind | intMask);
}

TreeNode *BitOrNode::_fold()
{
  if ((opnd0()->constantType() == Common::BoolKind) == 
      (opnd1()->constantType() == Common::BoolKind))
    return new PrimitiveLitNode(opnd0()->literal() | opnd1()->literal(), position());
  else return this;

}

Common::Kind BitXorNode::foldMask()
{
  return (Common::Kind)(Common::BoolKind | intMask);
}

TreeNode *BitXorNode::_fold()
{
  if ((opnd0()->constantType() == Common::BoolKind) == 
      (opnd1()->constantType() == Common::BoolKind))
    return new PrimitiveLitNode(opnd0()->literal() ^ opnd1()->literal(), position());
  else return this;

}

Common::Kind CandNode::foldMask()
{
  return Common::BoolKind;
}

TreeNode *CandNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal() && opnd1()->literal(), position());
}

Common::Kind CorNode::foldMask()
{
  return Common::BoolKind;
}

TreeNode *CorNode::_fold()
{
  return new PrimitiveLitNode(opnd0()->literal() || opnd1()->literal(), position());
}

static TreeNode *foldField(FieldAccessNode *f)
{
  Decl *d = f->decl();

  // Only simple and qualified names will have decl()'s at this point,
  // and only those names should be folded...
  if (d && (d->modifiers() & Common::Final)  && (d->category() == Decl::Field) &&
      !d->source()->absent())
    {
      TreeNode *source = d->source()->initExpr();
      
      if (source->constantType())
	{
#if 0
	  cout << "Field folding\n";
	  f->print(cout);
	  cout << "giving\n";
	  source->print(cout);
#endif
	  // make a new node so that the correct source position is retained
	  return new PrimitiveLitNode(source->literal(), f->position());
	}
    }
  return f;
}

// Don't fold ObjectFieldAccessNodes or SuperFieldAccessNodes

TreeNode *TypeFieldAccessNode::foldConstants()
{
  ftype()->foldConstants();
  return foldField(this);
}

TreeNode *ThisFieldAccessNode::foldConstants()
{
  return foldField(this);
}

/* Constant folding in templates. We don't have a real class yet, but
   we do have instantiated fields and methods (accessible via the Decl) */
void TemplateEnv::foldConstants(void)
{
  foreach (i, llist<instance *>, *instances)
    {
      Decl *inst = (*i)->d;

      if (DEBUG)
	cout << "Constant folding " << *inst->name() << "\n";

      inst->asType()->foldConstants();

      foriter (member, inst->environ()->allProperDecls(), EnvironIter)
	{
	  member->source()->foldConstants();
	  member->type()->foldConstants();
	}
    }
}

void TemplateEnv::foldFields(void)
{
  foreach (i, llist<instance *>, *instances)
    {
      Decl *inst = (*i)->d;

      if (DEBUG)
	cout << "Field folding " << *inst->name() << "\n";

      foriter (member, inst->environ()->allProperDecls(), EnvironIter)
	member->source()->foldFields();
    }
}
