/*  $Archive:: /Ti/AMUDP/amudp_reqrep.cpp                                 $
 *     $Date: Wed, 26 Sep 2001 01:45:46 -0700 $
 * $Revision: 1.2 $
 * Description: AMUDP Implementations of request/reply operations
 * Copyright 2000, Dan Bonachea <bonachea@cs.berkeley.edu>
 */
#include <assert.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>
#ifndef WIN32
  #include <sys/time.h>
  #include <unistd.h>
  #include <fcntl.h>
#endif

#include <amudp.h>
#include <amudp_internal.h>
#include "socket.h"

/* forward decls */
static int AMUDP_RequestGeneric(amudp_category_t category, 
                          ep_t request_endpoint, int reply_endpoint, handler_t handler, 
                          void *source_addr, int nbytes, uintptr_t dest_offset, 
                          int numargs, va_list argptr,
                          uint8_t systemType, uint8_t systemArg);
static int AMUDP_ReplyGeneric(amudp_category_t category, 
                          amudp_buf_t *requestbuf, handler_t handler, 
                          void *source_addr, int nbytes, uintptr_t dest_offset, 
                          int numargs, va_list argptr,
                          uint8_t systemType, uint8_t systemArg);
/*------------------------------------------------------------------------------------
 * Private helpers
 *------------------------------------------------------------------------------------ */
static int intpow(int val, int exp) {
  int retval = 1;
  int i;
  assert(exp >= 0);
  for (i = 0; i < exp; i++) retval *= val;
  return retval;
  }
/* ------------------------------------------------------------------------------------ */
#ifdef UFXP
  #define getMicrosecondTimeStamp() ufxp_getustime()
#else
  #ifdef WIN32
    int64_t getMicrosecondTimeStamp() {
      static int status = -1;
      static double multiplier;
      if (status == -1) { /*  first time run */
        LARGE_INTEGER freq;
        if (!QueryPerformanceFrequency(&freq)) status = 0; /*  don't have high-perf counter */
        else {
          multiplier = 1000000 / (double)freq.QuadPart;
          status = 1;
          }
        }
      if (status) { /*  we have a high-performance counter */
        LARGE_INTEGER count;
        QueryPerformanceCounter(&count);
        return (int64_t)(multiplier * count.QuadPart);
        }
      else { /*  no high-performance counter */
        /*  this is a millisecond-granularity timer that wraps every 50 days */
        return (GetTickCount() * 1000);
        }
      }
  /* #elif defined(__I386__) 
   * TODO: it would be nice to take advantage of the Pentium's "rdtsc" instruction,
   * which reads a fast counter incremented on each cycle. Unfortunately, that
   * requires a way to convert cycles to microseconds, and there doesn't appear to 
   * be a way to directly query the cycle speed
   */

  #else /* unknown processor - use generic UNIX call */
    int64_t getMicrosecondTimeStamp() {
      int64_t retval;
      struct timeval tv;
      if (gettimeofday(&tv, NULL)) {
        perror("gettimeofday");
        abort();
        }
      retval = ((int64_t)tv.tv_sec) * 1000000 + tv.tv_usec;
      return retval;
      }
  #endif
#endif
/* ------------------------------------------------------------------------------------ */
static int sendPacket(ep_t ep, void *packet, int packetlength, en_t destaddress) {
  assert(ep && packet && packetlength > 0);
  #if USE_TRUE_BULK_XFERS
    assert(packetlength <= AMUDP_MAXBULK_NETWORK_MSG);
  #else
    assert(packetlength <= AMUDP_MAX_NETWORK_MSG);
  #endif

  #if UFXP
    if (ufxp_send(packet, packetlength, &destaddress) != UFXP_OK) {
      AMUDP_RETURN_ERRFR(RESOURCE, sendPacket, "ufxp_send() failed");
      }
  #else
    #if DEBUG_VERBOSE
    { char temp[80];
      fprintf(stderr, "sending packet to (%s)\n", AMUDP_enStr(destaddress, temp)); fflush(stderr);
      }
    #endif
    if (sendto(ep->s, (char *)packet, packetlength, /* Solaris requires cast to char* */
               0, (struct sockaddr *)&destaddress, sizeof(en_t)) == SOCKET_ERROR) {
      AMUDP_RETURN_ERRFR(RESOURCE, sendPacket, sockErrDesc());
      }
  #endif
  ep->stats.TotalBytesSent += packetlength;
  return AM_OK;
  }
/* ------------------------------------------------------------------------------------ */
static int AMUDP_GetOpcode(int isrequest, amudp_category_t cat) {
  switch (cat) {
    case amudp_Short:
      if (isrequest) return AM_REQUEST_M;
      else return AM_REPLY_M;
    case amudp_Medium:
      if (isrequest) return AM_REQUEST_IM;
      else return AM_REPLY_IM;
    case amudp_Long:
      if (isrequest) return AM_REQUEST_XFER_M;
      else return AM_REPLY_XFER_M; 
    default: assert(0);
      return -1;
    }
  }
/* ------------------------------------------------------------------------------------ */
static int sourceAddrToId(ep_t ep, en_t sourceAddr) {
  /*  return source id in ep perproc table of this remote addr, or -1 for not found */
  int i; /*  TODO: make this more efficient */
  for (i = 0; i < ep->P; i++) {
    if (enEqual(ep->perProcInfo[i].remoteName, sourceAddr))
      return i;
    }
  return -1;
  }
/* ------------------------------------------------------------------------------------ */
/* these handle indexing into our 2-D array of desriptors and buffers */
#define GET_REQ_DESC(ep, remoteProc, inst) \
  (&((ep)->requestDesc[inst * (ep)->P + remoteProc]))
#define GET_REQ_BUF(ep, remoteProc, inst) \
  (&((ep)->requestBuf[inst * (ep)->P + remoteProc]))
#define GET_REP_DESC(ep, remoteProc, inst) \
  (&((ep)->replyDesc[inst * (ep)->P + remoteProc]))
#define GET_REP_BUF(ep, remoteProc, inst) \
  (&((ep)->replyBuf[inst * (ep)->P + remoteProc]))
/* these recover the processor and instance indexes from an entry offset into the 2-D arrays */
#define GET_REMOTEPROC_FROM_POS(ep, curpos) \
  (curpos % (ep)->P)
#define GET_INST_FROM_POS(ep, curpos) \
  (curpos / (ep)->P)
/* ------------------------------------------------------------------------------------ */
/* accessors for packet args, data and length
 * the only complication here is we want data to be double-word aligned, so we may add
 * an extra unused 4-byte argument to make sure the data lands on a double-word boundary
 */
#define HEADER_EVEN_WORDLENGTH  (((int)((&((amudp_buf_t *)NULL)->_Data)-1))%8==0?1:0)
#define ACTUAL_NUM_ARGS(pMsg) (AMUDP_MSG_NUMARGS(pMsg)%2==0?       \
                            AMUDP_MSG_NUMARGS(pMsg)+!HEADER_EVEN_WORDLENGTH:  \
                            AMUDP_MSG_NUMARGS(pMsg)+HEADER_EVEN_WORDLENGTH)

#define GET_PACKET_LENGTH(pbuf)                                       \
  (((char *)&pbuf->_Data[4*ACTUAL_NUM_ARGS(&pbuf->Msg) + pbuf->Msg.nBytes]) - ((char *)pbuf))
#define GET_PACKET_DATA(pbuf)                                         \
  (&pbuf->_Data[4*ACTUAL_NUM_ARGS(&pbuf->Msg)])
#define GET_PACKET_ARGS(pbuf)                                         \
  ((uint32_t *)pbuf->_Data)
/* ------------------------------------------------------------------------------------ */
#define RUN_HANDLER_SHORT(phandlerfn, token, pArgs, numargs) do {                       \
  assert(phandlerfn);                                                                   \
  if (numargs == 0) (*(AMUDP_HandlerShort)phandlerfn)((void *)token);                   \
  else {                                                                                \
    uint32_t *args = (uint32_t *)(pArgs); /* eval only once */                          \
    switch (numargs) {                                                                  \
      case 1: (*(AMUDP_HandlerShort)phandlerfn)((void *)token, args[0]); break;         \
      case 2: (*(AMUDP_HandlerShort)phandlerfn)((void *)token, args[0], args[1]); break;\
      case 3: (*(AMUDP_HandlerShort)phandlerfn)((void *)token, args[0], args[1], args[2]); break; \
      case 4: (*(AMUDP_HandlerShort)phandlerfn)((void *)token, args[0], args[1], args[2], args[3]); break; \
      case 5: (*(AMUDP_HandlerShort)phandlerfn)((void *)token, args[0], args[1], args[2], args[3], args[4]); break; \
      case 6: (*(AMUDP_HandlerShort)phandlerfn)((void *)token, args[0], args[1], args[2], args[3], args[4], args[5]); break; \
      case 7: (*(AMUDP_HandlerShort)phandlerfn)((void *)token, args[0], args[1], args[2], args[3], args[4], args[5], args[6]); break; \
      case 8: (*(AMUDP_HandlerShort)phandlerfn)((void *)token, args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7]); break; \
      default: abort();                                                                 \
      }                                                                                 \
    }                                                                                   \
  } while (0)
/* ------------------------------------------------------------------------------------ */
#define _RUN_HANDLER_MEDLONG(phandlerfn, token, pArgs, numargs, pData, datalen) do {   \
  assert(phandlerfn);                                                         \
  if (numargs == 0) (*phandlerfn)(token, pData, datalen);                     \
  else {                                                                      \
    uint32_t *args = (uint32_t *)(pArgs); /* eval only once */                \
    switch (numargs) {                                                        \
      case 1: (*phandlerfn)(token, pData, datalen, args[0]); break;           \
      case 2: (*phandlerfn)(token, pData, datalen, args[0], args[1]); break;  \
      case 3: (*phandlerfn)(token, pData, datalen, args[0], args[1], args[2]); break; \
      case 4: (*phandlerfn)(token, pData, datalen, args[0], args[1], args[2], args[3]); break; \
      case 5: (*phandlerfn)(token, pData, datalen, args[0], args[1], args[2], args[3], args[4]); break; \
      case 6: (*phandlerfn)(token, pData, datalen, args[0], args[1], args[2], args[3], args[4], args[5]); break; \
      case 7: (*phandlerfn)(token, pData, datalen, args[0], args[1], args[2], args[3], args[4], args[5], args[6]); break; \
      case 8: (*phandlerfn)(token, pData, datalen, args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7]); break; \
      default: abort();                                                                 \
      }                                                                                 \
    }                                                                                   \
  } while (0)
#define RUN_HANDLER_MEDIUM(phandlerfn, token, pArgs, numargs, pData, datalen) do {      \
    assert(((int)pData) % 8 == 0);  /* we guarantee double-word alignment for data payload of medium xfers */ \
    _RUN_HANDLER_MEDLONG((AMUDP_HandlerMedium)phandlerfn, (void *)token, pArgs, numargs, (void *)pData, (int)datalen); \
    } while(0)
#define RUN_HANDLER_LONG(phandlerfn, token, pArgs, numargs, pData, datalen)             \
  _RUN_HANDLER_MEDLONG((AMUDP_HandlerLong)phandlerfn, (void *)token, pArgs, numargs, (void *)pData, (int)datalen)
/* ------------------------------------------------------------------------------------ */
#ifdef UFXP
  #define AMUDP_DrainNetwork(ep) AM_OK  /*  already happens asynchronously for ufxp */

  static int AMUDP_WaitForEndpointActivity(eb_t eb, struct timeval *tv) {
    /* drain network and block up to tv time for endpoint recv buffers to become non-empty (NULL to block)
     * return AM_OK for activity, AM_ERR_ for other error, -1 for timeout 
     * wakeupOnControlActivity controls whether we return on control socket activity (for blocking)
     */
    ep_t ep = eb->endpoints[0];
    assert(eb->n_endpoints == 1);
    
    if (tv == NULL) { /* block indefinitely */
      /* we need to wake up and poll SPMD control periodically */
      while (1) {
        int waittime = 100000; /* every 100 milliseconds */
        int retval = ufxp_recv(NULL, NULL, waittime, NULL);    
        if (retval == UFXP_OK) return AM_OK;
        else if (retval == UFXP_ERR_TIMEDOUT) {
          int activity = 0;
          AMUDP_SPMDHandleControlTraffic(&activity);
          if (activity && AMUDP_SPMDwakeupOnControlActivity) return AM_OK;
          continue;
          }
        else AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_WaitForEndpointActivity, "ufxp_recv NULL wait failed");
        }
      }
    else {
      int retval;
      int waittime = tv->tv_sec * 1000000 + tv->tv_usec;
      retval = ufxp_recv(NULL, NULL, waittime, NULL);    
      if (retval == UFXP_OK) return AM_OK;
      else if (retval == UFXP_ERR_TIMEDOUT) return -1;
      else AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_WaitForEndpointActivity, "ufxp_recv NULL wait failed");
      }
    }
/* ------------------------------------------------------------------------------------ */
#else
/* ioctl UDP fiasco:
 * ioctlsocket(FIONREAD) on a SOCK_DGRAM should return 
 * the size of the next message waiting, not the total data
 * available on the socket. We need this to decide whether 
 * or not we have an incoming bulk message next on the queue
 * This works on Linux, but Win2K seems to fuck it up (despite the 
 * fact their own Winsock spec says it returns the next message size)
 */
#if defined(WIN32) 
  #define BROKEN_IOCTL 1
#elif defined(AIX)
  #define BROKEN_IOCTL 1 // AIX is broken too...
#else 
  #define BROKEN_IOCTL 0
#endif
/* ------------------------------------------------------------------------------------ */
  /*  AMUDP_DrainNetwork - read anything outstanding from hardware/kernel buffers into app space */
  static int AMUDP_DrainNetwork(ep_t ep) {
    int totalBytesDrained = 0;
    while (1) {
      unsigned long bytesAvail;
      if (ioctlsocket(ep->s, _FIONREAD, &bytesAvail) == SOCKET_ERROR)
        AMUDP_RETURN_ERRFR(RESOURCE, "ioctlsocket()", sockErrDesc());
      if (bytesAvail == 0) break; 

      #if BROKEN_IOCTL
        if (bytesAvail > AMUDP_MAX_NETWORK_MSG) { 
          /* this workaround is a HACK that lets us decide if we truly have a bulk message */
          static char *junk = NULL;
          int retval;
          if (!junk) junk = (char *)malloc(AMUDP_MAX_NETWORK_MSG);
          retval = recvfrom(ep->s, junk, AMUDP_MAX_NETWORK_MSG, MSG_PEEK, NULL, NULL);
          if (retval == SOCKET_ERROR && 
            #ifdef WIN32
              WSAGetLastError() != WSAEMSGSIZE)
            #else
              errno != EFAULT) /* AIX */
            #endif
            AMUDP_RETURN_ERRFR(RESOURCE, "recv(MSG_PEEK) - broken ioctl Hack", sockErrDesc());
          if (retval < (int)bytesAvail) bytesAvail = retval; /* the true next message size */
          }
      #endif

      /* something waiting, acquire a buffer for it */
      { amudp_buf_t *freebuf;
        amudp_buf_t *destbuf;
        int destbufsz = AMUDP_MAX_NETWORK_MSG;
        struct sockaddr sa;
        int retval;
        int sz = sizeof(en_t);
        if (((ep->rxFreeIdx + 1) % ep->rxNumBufs) == ep->rxReadyIdx) { 
          /* out of buffers - postpone draining */
          #ifdef DEBUG
            ErrMessage("Receive buffer full - unable to drain network (this is usually caused by retransmissions)");
          #endif
          break;
          }
        freebuf = &ep->rxBuf[ep->rxFreeIdx];
        #if USE_TRUE_BULK_XFERS
          #if !BROKEN_IOCTL
            /* can't so this check when ioctl is broken */
            if (bytesAvail > AMUDP_MAXBULK_NETWORK_MSG) {
              char x;
              int retval = recvfrom(ep->s, (char *)&x, 1, MSG_PEEK, NULL, NULL);
              fprintf(stderr, "bytesAvail=%i  recvfrom(MSG_PEEK)=%i\n", (int)bytesAvail, retval); fflush(stderr);
              AMUDP_RETURN_ERRFR(RESOURCE, "AMUDP_DrainNetwork: received message that was too long", sockErrDesc());
              }
          #endif
          if (bytesAvail > AMUDP_MAX_NETWORK_MSG) { /* this is a true bulk buffer */
            destbuf = AMUDP_AcquireBulkBuffer(ep);
            freebuf->status.bulkBuffer = destbuf;
            destbufsz = AMUDP_MAXBULK_NETWORK_MSG;
            }
          else destbuf = freebuf;
        #else
          destbuf = freebuf;
        #endif
        destbuf->status.bulkBuffer = NULL;

        retval = myrecvfrom(ep->s, (char *)destbuf, destbufsz, 0, 
                          &sa, &sz);

        if (retval == SOCKET_ERROR)
          AMUDP_RETURN_ERRFR(RESOURCE, "AMUDP_DrainNetwork: recvfrom()", sockErrDesc());
        else if (retval == 0)
          AMUDP_RETURN_ERRFR(RESOURCE, "AMUDP_DrainNetwork: recvfrom() returned zero", sockErrDesc());
        else if (retval < AMUDP_MIN_NETWORK_MSG) 
          AMUDP_RETURN_ERRFR(RESOURCE, "AMUDP_DrainNetwork: incomplete message received in recvfrom()", sockErrDesc());
        else if (retval > destbufsz) 
            AMUDP_RETURN_ERRFR(RESOURCE, "AMUDP_DrainNetwork: buffer overrun in recvfrom()", sockErrDesc());
        #if defined(DEBUG) && !BROKEN_IOCTL
        else if (retval != (int)bytesAvail) { /* detect other broken ioctl implementations */
          fprintf(stderr, "bytesAvail=%i  recvfrom returned:%i  ioctl() is probably broken\n", (int)bytesAvail, retval); fflush(stderr);
          }
        #endif
        totalBytesDrained += retval;
        if (sz != sizeof(en_t))
          AMUDP_RETURN_ERRFR(RESOURCE, "AMUDP_DrainNetwork: recvfrom() returned wrong sockaddr size", sockErrDesc());
        freebuf->status.sourceAddr = *(en_t *)&sa;
        freebuf->status.handlerRunning = FALSE;
        ep->rxFreeIdx = (ep->rxFreeIdx + 1) % ep->rxNumBufs; /* mark in use */
        }
      }
    #if !defined(UFXP) && USE_SOCKET_RECVBUFFER_GROW
      /* heuristically decide whether we should expand the OS socket recv buffers */
      if (totalBytesDrained + AMUDP_MAXBULK_NETWORK_MSG > ep->socketRecvBufferSize) {
        /* it's possible we dropped something due to insufficient OS socket buffer space */
        if (!ep->socketRecvBufferMaxedOut) { /* try to do something about it */
          /* TODO: we may want to add some hysterisis here to prevent artifical inflation
           * due to retransmits after a long period of no polling 
           */
          /*int newsize = ep->socketRecvBufferSize + AMUDP_MAXBULK_NETWORK_MSG; - too slow */
          int newsize = 2 * ep->socketRecvBufferSize;
          int sanitymax = AMUDP_RECVBUFFER_MAX;

          if (newsize > sanitymax) { /* create a semi-sane upper bound */
            AMUDP_growSocketRecvBufferSize(ep, sanitymax);
            ep->socketRecvBufferMaxedOut = 1;
            }
          else AMUDP_growSocketRecvBufferSize(ep, newsize);
          }
        }
    #endif
    return AM_OK; /* done */
    }
  static int AMUDP_WaitForEndpointActivity(eb_t eb, struct timeval *tv) {
    /* drain network and block up to tv time for endpoint recv buffers to become non-empty (NULL to block)
     * return AM_OK for activity, AM_ERR_ for other error, -1 for timeout 
     * wakeupOnControlActivity controls whether we return on control socket activity (for blocking)
     */
    struct timeval prvtv;
    if (tv) { /* get our own private copy of tv we can modify */
      memcpy(&prvtv, tv, sizeof(struct timeval));
      tv = &prvtv;
      }

    {/* drain network and see if some receive buffer already non-empty */
      int i;
      for (i = 0; i < eb->n_endpoints; i++) {
        ep_t ep = eb->endpoints[i];
        int retval = AMUDP_DrainNetwork(ep);
        if (retval != AM_OK) AMUDP_RETURN(retval);
        }
      for (i = 0; i < eb->n_endpoints; i++) {
        ep_t ep = eb->endpoints[i];
        if (ep->rxReadyIdx != ep->rxFreeIdx) return AM_OK;
        }
      }

    while (1) {
      fd_set sockset;
      fd_set* psockset = &sockset;
      int i;
      int maxfd = 0;
      int64_t starttime, endtime;

      FD_ZERO(psockset);
      for (i = 0; i < eb->n_endpoints; i++) {
        FD_SET(eb->endpoints[i]->s, psockset);
        if ((int)eb->endpoints[i]->s > maxfd) maxfd = eb->endpoints[i]->s;
        }
      if (AMUDP_SPMDControlSocket != INVALID_SOCKET) {
        ASYNC_TCP_DISABLE();
        FD_SET(AMUDP_SPMDControlSocket, psockset);
        if ((int)AMUDP_SPMDControlSocket > maxfd) maxfd = AMUDP_SPMDControlSocket;
        }
      /* wait for activity */
      starttime = getMicrosecondTimeStamp();
      { int retval = select(maxfd+1, psockset, NULL, NULL, tv);
        if (AMUDP_SPMDControlSocket != INVALID_SOCKET) ASYNC_TCP_ENABLE();
        if (retval == SOCKET_ERROR) { 
          AMUDP_RETURN_ERRFR(RESOURCE, "AMUDP_Block: select()", sockErrDesc());
          }
        else if (retval == 0) return -1; /* time limit expired */
        }
      endtime = getMicrosecondTimeStamp();
      if (FD_ISSET(AMUDP_SPMDControlSocket, psockset)) {
        AMUDP_SPMDIsActiveControlSocket = TRUE; /* we may have missed a signal */
        AMUDP_SPMDHandleControlTraffic(NULL);
        if (AMUDP_SPMDwakeupOnControlActivity) break;
        }
      else break; /* activity on some endpoint in bundle */

      if (tv) { /* readjust remaining time */
        int64_t remainingtime = ((int64_t)tv->tv_sec) * 1000000 + tv->tv_usec;
        remainingtime = remainingtime - (endtime - starttime);
        if (remainingtime < 0) return -1; /* time limit expired */
        tv->tv_sec = (long)(remainingtime / 1000000);
        tv->tv_usec = (long)(remainingtime % 1000000);
        }
      }

    return AM_OK; /* some endpoint activity is waiting */
    }
#endif
/* ------------------------------------------------------------------------------------ */
static int AMUDP_HandleRequestTimeouts(ep_t ep, int numtocheck) {
  /* check the next numtocheck requests for timeout (or -1 for all)
   * and retransmit as necessary. return AM_OK or AM_ERR_XXX
   */
  int curpos = ep->timeoutCheckPosn;
  int numdesc = ep->P * ep->depth;
  int64_t now = getMicrosecondTimeStamp();
  int i;
  if (numtocheck <= 0 || numtocheck > numdesc) numtocheck = numdesc;
  for (i = 0; i < numtocheck; i++) {
    if (ep->requestDesc[curpos].inuse && 
      ep->requestDesc[curpos].timestamp <= now) {
      amudp_buf_t *basicbuf = &ep->requestBuf[curpos];
      amudp_buf_t *outgoingbuf = 
        (basicbuf->status.bulkBuffer ? 
         basicbuf->status.bulkBuffer :
         basicbuf);
      amudp_bufstatus_t *outgoingstatus = &ep->requestBuf[curpos].status;
      amudp_bufdesc_t *outgoingdesc = &ep->requestDesc[curpos];
      int timetowait = AMUDP_INITIAL_REQUESTTIMEOUT_MICROSEC * 
           intpow(AMUDP_REQUESTTIMEOUT_BACKOFF_MULTIPLIER, outgoingdesc->retryCount);
      if (timetowait >= AMUDP_MAX_REQUESTTIMEOUT_MICROSEC) {
        /* we already waited too long - request is undeliverable */
        int isrequest = AMUDP_MSG_ISREQUEST(&outgoingbuf->Msg);
        amudp_category_t cat = AMUDP_MSG_CATEGORY(&outgoingbuf->Msg);
        AMUDP_HandlerReturned handlerfn = (AMUDP_HandlerReturned)ep->handler[0];
        int opcode = AMUDP_GetOpcode(isrequest, cat);
        int destP = GET_REMOTEPROC_FROM_POS(ep, curpos);
        int instance = GET_INST_FROM_POS(ep, curpos);

        /* pretend this is a bounced recv buffer */
        /* note that source/dest for returned mesgs reflect the virtual "message denied" packet 
         * although it doesn't really matter because the AM2 spec is too vague
         * about the argblock returned message argument for it to be of any use to anyone
         */
        outgoingstatus->sourceId = (uint8_t)destP; 
        outgoingstatus->sourceAddr = ep->perProcInfo[destP].remoteName;
        outgoingstatus->dest = ep;

        outgoingstatus->replyIssued = TRUE; /* prevent any reply */
        outgoingstatus->handlerRunning = TRUE;
        assert(handlerfn);
        (*handlerfn)(ECONGESTION, opcode, (void *)outgoingbuf);
        outgoingstatus->handlerRunning = FALSE;

        outgoingdesc->inuse = FALSE; /* free it */        
        if (basicbuf->status.bulkBuffer) {
          AMUDP_ReleaseBulkBuffer(ep, basicbuf->status.bulkBuffer);
          basicbuf->status.bulkBuffer = NULL;
          }
        ep->perProcInfo[destP].instanceHint = (uint16_t)instance;
        ep->stats.ReturnedMessages++;
        }
      else {
        timetowait = timetowait * AMUDP_REQUESTTIMEOUT_BACKOFF_MULTIPLIER;
        outgoingdesc->retryCount++;
      
        /* retransmit */
        { /*  perform the send */
          int destP = GET_REMOTEPROC_FROM_POS(ep, curpos);
          int packetlength = GET_PACKET_LENGTH(outgoingbuf);
          en_t destaddress = ep->perProcInfo[destP].remoteName;
          /* tag should NOT be changed for retransmit */
          int retval;
          #if DEBUG_VERBOSE
            fprintf(stderr, "Retransmitting a request..."); fflush(stderr);
          #endif
          retval = sendPacket(ep, outgoingbuf, packetlength, destaddress);
          if (retval != AM_OK) AMUDP_RETURN(retval);        
          outgoingdesc->timestamp = getMicrosecondTimeStamp() + timetowait;
          ep->stats.RequestsRetransmitted[AMUDP_MSG_CATEGORY(&outgoingbuf->Msg)]++;
          }

        }
      }
    curpos = (curpos + 1) % numdesc;
    }
  
  /* advance checked posn */
  assert(curpos >= 0 && curpos < numdesc);
  ep->timeoutCheckPosn = curpos;

  return AM_OK;
  }
/* ------------------------------------------------------------------------------------ */
#define MAXINT64    ((((uint64_t)1) << 63) - 1)
#define MAXUINT64   ((uint64_t)-1)
static int64_t AMUDP_FindEarliestRequestTimeout(eb_t eb) {
  /* return the soonest timeout value for an active request
   * (which may have already passed)
   * return 0 for no outstanding requests
   */
  int64_t earliesttime = MAXINT64;
  int i;
  for (i = 0; i < eb->n_endpoints; i++) {
    ep_t ep = eb->endpoints[i];
    int numdesc = ep->P * ep->depth;
    int j;
    for (j = 0; j < numdesc; j++) {
      if (ep->requestDesc[j].inuse) {
        int64_t timestamp = ep->requestDesc[j].timestamp;
        if (timestamp < earliesttime) earliesttime = timestamp;
        }
      }
    }
  if (earliesttime == MAXINT64) return 0;
  else return earliesttime;
  }
/* ------------------------------------------------------------------------------------ */
extern int AMUDP_Block(eb_t eb) {
  /* block until some endpoint receive buffer becomes non-empty
   * does not poll, but does handle SPMD control socket events
   */

  /* first, quickly determine if something is already waiting */
  { struct timeval tv = {0,0};
    int retval = AMUDP_WaitForEndpointActivity(eb, &tv);
    if (retval != -1) AMUDP_RETURN(retval); /* error or something waiting */
    }

  while (1) {
    /* we need to be careful we don't sleep longer than the next packet timeout */
    int64_t nexttimeout = AMUDP_FindEarliestRequestTimeout(eb);
    int retval;
    if (nexttimeout) {
      struct timeval tv;
      int64_t now = getMicrosecondTimeStamp();
      if (nexttimeout < now) goto timeout; /* already have a request timeout */
      tv.tv_sec = (long)((nexttimeout - now) / 1000000);
      tv.tv_usec = (long)((nexttimeout - now) % 1000000);
      retval = AMUDP_WaitForEndpointActivity(eb, &tv);
      }
    else /* no outstanding requests, so just block */
      retval = AMUDP_WaitForEndpointActivity(eb, NULL); 
    if (retval != -1) AMUDP_RETURN(retval); /* error or something waiting */
     
    /* some request has timed out - handle it */
    timeout:
    { int i;
      for (i = 0; i < eb->n_endpoints; i++) {
        ep_t ep = eb->endpoints[i];
        int retval = AMUDP_HandleRequestTimeouts(ep, -1);
        if (retval != AM_OK) AMUDP_RETURN(retval);
        }
      }
    }

  }
/* ------------------------------------------------------------------------------------ */
#ifdef DEBUG
  #define REFUSE_NOTICE(reason) ErrMessage("I just refused a message and returned to sender. Reason: %s", reason)
#else
  #define REFUSE_NOTICE(reason) (void)0
#endif

/* this is a local-use-only macro for AMUDP_ServiceIncomingMessages */
#define AMUDP_REFUSEMESSAGE(ep, basicbuf, errcode) do {                         \
    amudp_buf_t *buf = ((basicbuf)->status.bulkBuffer ? (basicbuf)->status.bulkBuffer : (basicbuf));\
    int retval;                                                                 \
    buf->Msg.systemMessageType = (uint8_t)amudp_system_returnedmessage;         \
    buf->Msg.systemMessageArg = (uint8_t)errcode;                               \
    retval = sendPacket(ep, buf, GET_PACKET_LENGTH(buf), (basicbuf)->status.sourceAddr); \
       /* ignore errors sending this */                                         \
    if (retval != AM_OK) ErrMessage("failed to sendPacket to refuse message");  \
    else REFUSE_NOTICE(#errcode);                                               \
    goto donewithmessage;                                                       \
    } while(0)

/* main message receive workhorse - 
 * drain network once and service available incoming messages, up to AMUDP_MAX_RECVMSGS_PER_POLL
 */
static int AMUDP_ServiceIncomingMessages(ep_t ep) {
  /* drain network */
  int i;
  int retval = AMUDP_DrainNetwork(ep);
  if (retval != AM_OK) AMUDP_RETURN(retval);

  for (i = 0; AMUDP_MAX_RECVMSGS_PER_POLL == 0 || i < AMUDP_MAX_RECVMSGS_PER_POLL; i++) {
    int recvPoolNonempty = FALSE;
    #ifdef UFXP
    { int retval = ufxp_recv(NULL, NULL, 0, NULL);   
      if (retval == UFXP_OK) recvPoolNonempty = TRUE;
      else if (retval == UFXP_ERR_TIMEDOUT) recvPoolNonempty = FALSE;
      else AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_ServiceIncomingMessages, "ufxp_recv NULL wait failed");
      }
    #else
      recvPoolNonempty = (ep->rxReadyIdx != ep->rxFreeIdx);
    #endif
    if (!recvPoolNonempty) return AM_OK; /* nothing else waiting */
  
    /* we have a real message waiting - get it */
    { amudp_buf_t *basicbuf; /* the true buffer in the recv queue that holds status bits */
      amudp_buf_t *buf; /* the true buffer or bulk buffer that holds the msg */
      amudp_bufstatus_t* status; /* the status block for this buffer */
      amudp_msg_t *msg;
      #ifdef UFXP
      { int buflen;
        en_t sourceAddr;

        int retval = ufxp_recv((void *)&basicbuf, &buflen, 0, &sourceAddr); 
        buf = basicbuf; // TODO: support true bulk on UFXP

        #if DEBUG_VERBOSE
        { char temp[80];
          printf("ufxp_recv returned %i buflen=%i sourceAddr=%s\n", 
            retval, buflen, AMUDP_enStr(sourceAddr, temp));
          fflush(stdout);
          }
        #endif

        if (retval != UFXP_OK) 
          AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_ServiceIncomingMessages, "ufxp_recv timed out after a successful NULL wait");
        if (buflen > AMUDP_MAX_NETWORK_MSG)
          AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_ServiceIncomingMessages, "buffer overrun - received message too long");
        else if (buflen < AMUDP_MIN_NETWORK_MSG) 
          AMUDP_RETURN_ERRFR(RESOURCE, "AMUDP_DrainNetwork: incomplete message received in ufxp_recv()", sockErrDesc());

        basicbuf->status.sourceAddr = sourceAddr;
        }
      #else
        assert(ep->rxReadyIdx < ep->rxNumBufs);
        assert(ep->rxFreeIdx < ep->rxNumBufs);
        assert(ep->rxReadyIdx != ep->rxFreeIdx);
        basicbuf = &ep->rxBuf[ep->rxReadyIdx];
        buf = (basicbuf->status.bulkBuffer ? basicbuf->status.bulkBuffer : basicbuf);
        if (basicbuf->status.handlerRunning) {
          AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_ServiceIncomingMessages, "user caused a poll to occur while handler on the same bundle was running");
          }
      #endif

      status = &basicbuf->status;
      status->dest = ep; /* remember which ep recvd this message */
      msg = &buf->Msg;

      if (AMUDP_FaultInjectionEnabled) { /* allow fault injection to drop some revcd messages */
        double randval = rand() / (double)RAND_MAX;
        assert(randval >= 0.0 && AMUDP_FaultInjectionRate >= 0.0);
        if (randval < AMUDP_FaultInjectionRate) {
          #if DEBUG_VERBOSE
            fprintf(stderr, "fault injection dropping a packet..\n"); fflush(stderr);
          #endif
          goto donewithmessage;
          }
        }

      /* perform acceptance checks */
      { int numargs = AMUDP_MSG_NUMARGS(msg);
        int seqnum = AMUDP_MSG_SEQNUM(msg);
        int isrequest = AMUDP_MSG_ISREQUEST(msg);
        amudp_category_t cat = AMUDP_MSG_CATEGORY(msg);
        int issystemmsg = ((amudp_system_messagetype_t)msg->systemMessageType) != amudp_system_user;

        /* handle returned messages */
        if (issystemmsg) { 
          amudp_system_messagetype_t type = ((amudp_system_messagetype_t)msg->systemMessageType);
          if (type == amudp_system_returnedmessage) { 
            AMUDP_HandlerReturned handlerfn = (AMUDP_HandlerReturned)ep->handler[0];
            op_t opcode;
            int sourceID = sourceAddrToId(ep, buf->status.sourceAddr);
            if (sourceID < 0) goto donewithmessage; /*  unknown source, ignore message */
            status->sourceId = (uint8_t)sourceID;
            if (isrequest) { /*  the returned message is a request, so free that request buffer */
              amudp_bufdesc_t *desc = GET_REQ_DESC(ep, status->sourceId, msg->instance);
              amudp_buf_t *basicreqbuf = GET_REQ_BUF(ep, status->sourceId, msg->instance);
              if (desc->inuse && desc->seqNum == seqnum) {
                desc->inuse = FALSE;
                if (basicreqbuf->status.bulkBuffer) {
                  AMUDP_ReleaseBulkBuffer(ep, basicreqbuf->status.bulkBuffer);
                  basicreqbuf->status.bulkBuffer = NULL;
                  }
                desc->seqNum = (uint8_t)!(desc->seqNum);
                ep->perProcInfo[status->sourceId].instanceHint = msg->instance;
                }
              }
            opcode = AMUDP_GetOpcode(isrequest, cat);

            /* note that source/dest for returned mesgs reflect the virtual "message denied" packet 
             * although it doesn't really matter because the AM2 spec is too vague
             * about the argblock returned message argument for it to be of any use to anyone
             */
            status->replyIssued = TRUE; /* prevent any reply */
            status->handlerRunning = TRUE;
              assert(handlerfn);
              (*handlerfn)(msg->systemMessageArg, opcode, (void *)basicbuf);
            status->handlerRunning = FALSE;
            ep->stats.ReturnedMessages++;
            goto donewithmessage;
            }
          }

        if (isrequest) ep->stats.RequestsReceived[cat]++;
        else ep->stats.RepliesReceived[cat]++;

        if (ep->tag == AM_NONE || 
           (ep->tag != msg->tag && ep->tag != AM_ALL))
            AMUDP_REFUSEMESSAGE(ep, basicbuf, EBADTAG);
        if (msg->instance >= ep->depth)
            AMUDP_REFUSEMESSAGE(ep, basicbuf, EUNREACHABLE);
        if (ep->handler[msg->handlerId] == amudp_unused_handler &&
            !issystemmsg && msg->handlerId != 0)
            AMUDP_REFUSEMESSAGE(ep, basicbuf, EBADHANDLER);
      
        switch (cat) {
          case amudp_Short:
            if (msg->nBytes > 0 || msg->destOffset > 0)
              AMUDP_REFUSEMESSAGE(ep, basicbuf, EBADLENGTH);
            break;
          case amudp_Medium:
            if (msg->nBytes > AMUDP_MAX_MEDIUM || msg->destOffset > 0)
              AMUDP_REFUSEMESSAGE(ep, basicbuf, EBADLENGTH);
            break;
          case amudp_Long: 
            /* check segment limits */
            #if USE_TRUE_BULK_XFERS
              if (msg->nBytes > AMUDP_MAX_LONG)
                AMUDP_REFUSEMESSAGE(ep, basicbuf, EBADLENGTH);
            #else
              if (msg->nBytes > AMUDP_MAX_MEDIUM)
                AMUDP_REFUSEMESSAGE(ep, basicbuf, EBADLENGTH);
            #endif
            if (ep->segAddr == 0 || ep->segLength == 0 || msg->destOffset > AMUDP_MAX_SEGLENGTH)
              AMUDP_REFUSEMESSAGE(ep, basicbuf, EBADSEGOFF);
            if (msg->destOffset + msg->nBytes > ep->segLength || msg->nBytes > AMUDP_MAX_LONG)
              AMUDP_REFUSEMESSAGE(ep, basicbuf, EBADLENGTH);
            break;
          default:
            assert(0);
          }

        { /*  find the source id */
          int sourceID = sourceAddrToId(ep, status->sourceAddr);
          if (sourceID < 0) AMUDP_REFUSEMESSAGE(ep, basicbuf, EBADENDPOINT);
          status->sourceId = (uint8_t)sourceID;
          }

        /* check sequence number to see if this is a new request/reply or a duplicate */
        if (isrequest) {
          amudp_bufdesc_t *desc = GET_REP_DESC(ep, status->sourceId, msg->instance);
          if (seqnum != desc->seqNum) { 
            /*  request resent or reply got dropped - resend reply */
            amudp_buf_t *replybuf = GET_REP_BUF(ep, status->sourceId, msg->instance);
            int retval;
            assert(replybuf);
            if (replybuf->status.bulkBuffer) replybuf = replybuf->status.bulkBuffer;
            #if DEBUG_VERBOSE
              ErrMessage("Got a duplicate request - resending previous reply.");
            #endif
            retval = sendPacket(ep, replybuf, GET_PACKET_LENGTH(replybuf),
              ep->perProcInfo[status->sourceId].remoteName);
            if (retval != AM_OK) ErrMessage("sendPacket failed while resending a reply");
            ep->stats.RepliesRetransmitted[AMUDP_MSG_CATEGORY(&replybuf->Msg)]++;
            /*  ignore error return */
            goto donewithmessage;
            }
          }
        else {
          amudp_bufdesc_t *desc = GET_REQ_DESC(ep, status->sourceId, msg->instance);
          if (seqnum != desc->seqNum) { /*  duplicate reply, we already ran handler - ignore it */
            #if DEBUG_VERBOSE
              ErrMessage("Ignoring a duplicate reply.");
            #endif
            goto donewithmessage;
            }
          }

        /* --- message accepted --- */

        if (!isrequest) { /* it's a reply, free the corresponding request */
          int sourceId = status->sourceId;
          amudp_bufdesc_t *desc = GET_REQ_DESC(ep, sourceId, msg->instance);
          if (desc->inuse) { 
            amudp_buf_t *basicreqbuf = GET_REQ_BUF(ep, sourceId, msg->instance);
            desc->inuse = FALSE;
            if (basicreqbuf->status.bulkBuffer) {
              AMUDP_ReleaseBulkBuffer(ep, basicreqbuf->status.bulkBuffer);
              basicreqbuf->status.bulkBuffer = NULL;
              }
            desc->seqNum = (uint8_t)!(desc->seqNum); 
            ep->perProcInfo[sourceId].instanceHint = msg->instance;
            #if AMUDP_COLLECT_LATENCY_STATS
              { /* gather some latency statistics */
                uint64_t now = getMicrosecondTimeStamp();
                uint64_t latency = (now - desc->firstSendTime);
                ep->stats.RequestSumLatency += latency;
                if (latency < ep->stats.RequestMinLatency) ep->stats.RequestMinLatency = latency;
                if (latency > ep->stats.RequestMaxLatency) ep->stats.RequestMaxLatency = latency;
                }
            #endif
            }
          else { /* request timed out and we decided it was undeliverable, then a reply arrived */
            desc->seqNum = (uint8_t)!(desc->seqNum); /* toggle the seq num */
            /* TODO: seq numbers may get out of sync on timeout 
             * if request got through but replies got lost 
             * we also may do bad things if a reply to an undeliverable message 
             * arrives after we've reused the request buffer (very unlikely)
             * possible soln: add an epoch number
             */
            goto donewithmessage; /* reply handler should NOT be run in this situation */
            }
          }

        { /*  run the handler */
          status->replyIssued = FALSE;
          status->handlerRunning = TRUE;
          if (issystemmsg) { /* an AMUDP system message */
            amudp_system_messagetype_t type = ((amudp_system_messagetype_t)(msg->systemMessageType & 0xF));
            switch (type) {
              case amudp_system_autoreply:
                /*  do nothing, already taken care of */
                break;
              case amudp_system_bulkxferfragment:
                /*  perform bulk copy of fragment */
                memcpy(((int8_t *)ep->segAddr) + msg->destOffset, GET_PACKET_DATA(buf), msg->nBytes);
                { /* update our slot info and see if bulk xfer is complete */
                  int slotnum = msg->systemMessageType >> 4;
                  bulkslot_t *slot = &ep->perProcInfo[status->sourceId].inboundBulkSlot[slotnum];
                  if (slot->packetsRemaining == 0) {
                    /*  this is the first packet of bulk xfer */
                    assert(msg->systemMessageArg > 0);
                    slot->packetsRemaining = msg->systemMessageArg;
                    slot->minDestOffset = msg->destOffset;
                    slot->runningLength = msg->nBytes;
                    }
                  else { /*  not first, but possibly last  */
                    assert(slot->packetsRemaining <= msg->systemMessageArg);
                    slot->packetsRemaining--;
                    if (msg->destOffset < slot->minDestOffset) slot->minDestOffset = msg->destOffset;
                    slot->runningLength += msg->nBytes;
                    if (slot->packetsRemaining == 0) { /*  just processed last message, now run handler */
                      /*  mark it as a user message before running handler */
                      buf->Msg.systemMessageType = amudp_system_user; 
                      buf->Msg.systemMessageArg = 0;
                      
                      RUN_HANDLER_LONG(ep->handler[msg->handlerId], basicbuf,
                        GET_PACKET_ARGS(buf), numargs,  
                        (((int8_t *)ep->segAddr) + slot->minDestOffset), slot->runningLength);
                      }
                    }
                  }
                break;
              default:
                assert(0);
              }
            }
          else { /* a user message */
            switch (cat) {
              case amudp_Short: 
                RUN_HANDLER_SHORT(ep->handler[msg->handlerId], basicbuf, 
                                  GET_PACKET_ARGS(buf), numargs);
                break;
              case amudp_Medium: 
                RUN_HANDLER_MEDIUM(ep->handler[msg->handlerId], basicbuf, 
                                   GET_PACKET_ARGS(buf), numargs, 
                                   GET_PACKET_DATA(buf), msg->nBytes);
                break;
              case amudp_Long: {
                int8_t *pData = ((int8_t *)ep->segAddr) + msg->destOffset;
                /*  a single-message bulk transfer. do the copy */
                memcpy(pData, GET_PACKET_DATA(buf), msg->nBytes);
                RUN_HANDLER_LONG(ep->handler[msg->handlerId], basicbuf, 
                                   GET_PACKET_ARGS(buf), numargs, 
                                   pData, msg->nBytes);
                break;
                }
              default:
                assert(0);
              }
            }
          status->handlerRunning = FALSE;
          if (isrequest && !status->replyIssued) {
            /*  user didn't reply, so issue an auto-reply */
            if (AMUDP_ReplyGeneric(amudp_Short, buf, 0, 0, 0, 0, 0, NULL, amudp_system_autoreply, 0) 
                != AM_OK) /*  should never happen - don't return here to prevent leaking buffer */
              ErrMessage("Failed to issue auto reply in AMUDP_ServiceIncomingMessages");
            }
          if (isrequest) { /*  message was a request, alternate the reply sequence number so duplicates of this request get ignored */
            amudp_bufdesc_t *desc = GET_REP_DESC(ep, status->sourceId, msg->instance);
            desc->seqNum = (uint8_t)!(desc->seqNum);
            }
          }
        } /*  acceptance checks */

      donewithmessage: /* message handled - continue to next one */

      /* free the handled buffer */
      if (status->bulkBuffer) {
        AMUDP_ReleaseBulkBuffer(ep, status->bulkBuffer);
        status->bulkBuffer = NULL;
        }
      #ifdef UFXP
        { int retval;
          retval = ufxp_freerxbuf(buf);
          if (retval != UFXP_OK) AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_ServiceIncomingMessages, "ufxp_freerxbuf failed");
          }
      #else
        assert(ep->rxReadyIdx < ep->rxNumBufs);
        assert(ep->rxFreeIdx < ep->rxNumBufs);
        assert(ep->rxReadyIdx != ep->rxFreeIdx);
        ep->rxReadyIdx = (ep->rxReadyIdx + 1) % ep->rxNumBufs; /* remove from queue and put back on free list */
      #endif

      } /*  message waiting */
    }  /*  for */
  return AM_OK;
  } /*  AMUDP_ServiceIncomingMessages */
#undef AMUDP_REFUSEMESSAGE  /* this is a local-use-only macro */
/*------------------------------------------------------------------------------------
 * Poll
 *------------------------------------------------------------------------------------ */
extern int AM_Poll(eb_t eb) {
  int i;
  AMUDP_CHECKINIT();
  if (!eb) AMUDP_RETURN_ERR(BAD_ARG);

  for (i = 0; i < eb->n_endpoints; i++) {
    int retval;
    ep_t ep = eb->endpoints[i];

    if (ep->depth != -1) { /* only poll endpoints which have buffers */

      retval = AMUDP_SPMDHandleControlTraffic(NULL);
      if (retval != AM_OK) AMUDP_RETURN(retval);

      retval = AMUDP_ServiceIncomingMessages(ep); /* drain network and check for activity */
      if (retval != AM_OK) AMUDP_RETURN(retval);

      retval = AMUDP_HandleRequestTimeouts(ep, AMUDP_TIMEOUTS_CHECKED_EACH_POLL);
      if (retval != AM_OK) AMUDP_RETURN(retval);

      }
    }

  return AM_OK;
  }
/*------------------------------------------------------------------------------------
 * Generic Request/Reply
 *------------------------------------------------------------------------------------ */
static int AMUDP_RequestGeneric(amudp_category_t category, 
                          ep_t request_endpoint, int reply_endpoint, handler_t handler, 
                          void *source_addr, int nbytes, uintptr_t dest_offset, 
                          int numargs, va_list argptr, 
                          uint8_t systemType, uint8_t systemArg) {
  int destP;
  int instance;
  amudp_buf_t *basicbuf;
  amudp_buf_t *outgoingbuf;
  amudp_bufdesc_t *outgoingdesc;
  int packetlength;

  /*  always poll before sending a request */
  AM_Poll(request_endpoint->eb);

  destP = request_endpoint->translation[reply_endpoint].id;

  /*  acquire a free request buffer */
  {
    int depth = request_endpoint->depth;
    int found = FALSE;
    while (!found) {
      int hint = request_endpoint->perProcInfo[destP].instanceHint;
      if (!GET_REQ_DESC(request_endpoint, destP, hint)->inuse) { /*  hint is right */
        instance = hint;
        request_endpoint->perProcInfo[destP].instanceHint = (uint16_t)((hint+1)%depth);
        found = TRUE;
        }
      else { /*  hint is wrong */
        /*  search for a free instance */
        for (instance = (hint+1)%depth; instance != hint; instance = (instance+1)%depth) {
          if (!GET_REQ_DESC(request_endpoint, destP, instance)->inuse) {
            found = TRUE;
            break;
            }
          }
        if (!found) { 
          /*  no buffers available - wait until one is open 
           *  (hint will point to a free buffer) 
           *  TODO: we may consider some spin polling here
           */
          do {
            int retval;
            retval = AMUDP_Block(request_endpoint->eb);
            if (retval != AM_OK) AMUDP_RETURN(retval);
            retval = AM_Poll(request_endpoint->eb);
            if (retval != AM_OK) AMUDP_RETURN(retval);
            hint = request_endpoint->perProcInfo[destP].instanceHint;
            } while (GET_REQ_DESC(request_endpoint, destP, hint)->inuse);
          }
        }
      }
    }
  basicbuf = GET_REQ_BUF(request_endpoint, destP, instance);
  outgoingdesc = GET_REQ_DESC(request_endpoint, destP, instance);
  assert(!outgoingdesc->inuse);
  outgoingdesc->inuse = TRUE; /*  grab it now to claim as ours */

  if (nbytes > AMUDP_MAX_MEDIUM) {
    assert(category == amudp_Long && USE_TRUE_BULK_XFERS);
    outgoingbuf = AMUDP_AcquireBulkBuffer(request_endpoint);
    basicbuf->status.bulkBuffer = outgoingbuf;
    }
  else {
    basicbuf->status.bulkBuffer = NULL;
    outgoingbuf = basicbuf;
    }

  /*  setup message meta-data */
  { amudp_msg_t *msg = &outgoingbuf->Msg;
    AMUDP_MSG_SETFLAGS(msg, TRUE, category, numargs, outgoingdesc->seqNum);
    msg->destOffset = dest_offset;
    msg->handlerId = handler;
    msg->instance = (uint16_t)instance;
    msg->nBytes = (uint16_t)nbytes;
    msg->systemMessageType = systemType;
    msg->systemMessageArg = systemArg;
    }

  { /*  setup args */
    int i;
    uint32_t *args = GET_PACKET_ARGS(outgoingbuf);
    for (i = 0; i < numargs; i++) {
      args[i] = (uint32_t)va_arg(argptr, int); /* must be int due to default argument promotion */
      }
    }

  { /*  setup data */
    if (nbytes > 0) {
      memcpy(GET_PACKET_DATA(outgoingbuf), source_addr, nbytes);
      }
    }

  { /*  perform the send */
    en_t destaddress = request_endpoint->translation[reply_endpoint].name;
    int retval;
    packetlength = GET_PACKET_LENGTH(outgoingbuf);
    outgoingbuf->Msg.tag = request_endpoint->translation[reply_endpoint].tag;
    retval = sendPacket(request_endpoint, outgoingbuf, packetlength, destaddress);
    if (retval != AM_OK) {
      outgoingdesc->inuse = FALSE; /*  send failed, so message rejected - release buffer */
      if (basicbuf->status.bulkBuffer) {
        AMUDP_ReleaseBulkBuffer(request_endpoint, basicbuf->status.bulkBuffer);
        basicbuf->status.bulkBuffer = NULL;
        }
      request_endpoint->perProcInfo[reply_endpoint].instanceHint = (uint16_t)instance;
      AMUDP_RETURN(retval);
      }
    }

  /* outgoingdesc->seqNum = !(outgoingdesc->seqNum); */ /* this gets handled by AMUDP_ServiceIncomingMessages */
  { uint64_t now = getMicrosecondTimeStamp();
    uint32_t ustimeout = AMUDP_INITIAL_REQUESTTIMEOUT_MICROSEC;
    uint32_t expectedtime = (uint32_t)((2 * 1000000.0 / 1024.0) * packetlength / AMUDP_ExpectedBandwidth);
    /* bulk transfers may have a noticeable wire delay, so we grow the initial timeout
     * accordingly to allow time for the transfer to arrive and be serviced
     * These are the transfers that are really expensive to retransmit, 
     * so we want to avoid that until we're relatively certain they've really been lost
     */
    outgoingdesc->retryCount = 0;
    while (ustimeout < expectedtime &&
      ustimeout < AMUDP_MAX_REQUESTTIMEOUT_MICROSEC) {
      outgoingdesc->retryCount++;
      ustimeout *= AMUDP_REQUESTTIMEOUT_BACKOFF_MULTIPLIER;
      }
    outgoingdesc->timestamp = now + ustimeout;
    request_endpoint->stats.RequestsSent[category]++;
    #if AMUDP_COLLECT_LATENCY_STATS
      outgoingdesc->firstSendTime = now;
    #endif
    request_endpoint->stats.DataBytesSent[category] += sizeof(int) * numargs + nbytes;
    }
  return AM_OK;
  }
/* ------------------------------------------------------------------------------------ */
static int AMUDP_ReplyGeneric(amudp_category_t category, 
                          amudp_buf_t *requestbuf, handler_t handler, 
                          void *source_addr, int nbytes, uintptr_t dest_offset, 
                          int numargs, va_list argptr,
                          uint8_t systemType, uint8_t systemArg) {
  ep_t ep;
  int destP;
  int instance;
  amudp_buf_t *requestbasicbuf;
  amudp_buf_t *basicbuf;
  amudp_buf_t *outgoingbuf;
  amudp_bufdesc_t *outgoingdesc;

  /*  we don't poll within a reply because by definition we are already polling somewhere in the call chain */

  requestbasicbuf = requestbuf;
  requestbuf = (requestbasicbuf->status.bulkBuffer ? requestbasicbuf->status.bulkBuffer : requestbasicbuf);
  destP = requestbasicbuf->status.sourceId;
  ep = requestbasicbuf->status.dest;

  /*  acquire a free buffer  */
  /*  trivial because replies can safely overwrite previous reply in request instance */
  instance = requestbuf->Msg.instance; 

  basicbuf = GET_REP_BUF(ep, destP, instance);
  outgoingdesc = GET_REP_DESC(ep, destP, instance);

  if (basicbuf->status.bulkBuffer) { /* free bulk buffer of previous reply */
    AMUDP_ReleaseBulkBuffer(ep, basicbuf->status.bulkBuffer);
    basicbuf->status.bulkBuffer = NULL;
    }

  if (nbytes > AMUDP_MAX_MEDIUM) {
    assert(category == amudp_Long && USE_TRUE_BULK_XFERS);
    outgoingbuf = AMUDP_AcquireBulkBuffer(ep);
    basicbuf->status.bulkBuffer = outgoingbuf;
    }
  else {
    basicbuf->status.bulkBuffer = NULL;
    outgoingbuf = basicbuf;
    }

  /*  setup message meta-data */
  { amudp_msg_t *msg = &outgoingbuf->Msg;
    AMUDP_MSG_SETFLAGS(msg, FALSE, category, numargs, outgoingdesc->seqNum);
    msg->destOffset = dest_offset;
    msg->handlerId = handler;
    msg->instance = (uint16_t)instance;
    msg->nBytes = (uint16_t)nbytes;
    msg->systemMessageType = systemType;
    msg->systemMessageArg = systemArg;
    }

  { /*  setup args */
    int i;
    uint32_t *args = GET_PACKET_ARGS(outgoingbuf);
    for (i = 0; i < numargs; i++) {
      args[i] = (uint32_t)va_arg(argptr, int); /* must be int due to default argument promotion */
      }
    #if USE_CLEAR_UNUSED_SPACE
      for ( ; i < AMUDP_MAX_SHORT; i++) {
        args[i] = 0;
        }
    #endif
    }

  { /*  setup data */
    memcpy(GET_PACKET_DATA(outgoingbuf), source_addr, nbytes);
    #if 0 /* not necessary- we never send this stuff */
      #if USE_CLEAR_UNUSED_SPACE
        memset(&(GET_PACKET_DATA(outgoingbuf)[nbytes]), 0, AMUDP_MAX_LONG - nbytes);
      #endif
    #endif
    }

  { /*  perform the send */
    int packetlength = GET_PACKET_LENGTH(outgoingbuf);
    int destId = requestbasicbuf->status.sourceId;
    en_t destaddress = ep->perProcInfo[destId].remoteName;
    int retval;
    outgoingbuf->Msg.tag = ep->perProcInfo[destId].tag;
    retval = sendPacket(ep, outgoingbuf, packetlength, destaddress);
    if (retval != AM_OK) AMUDP_RETURN(retval);
    }

  /* outgoingdesc->seqNum = !(outgoingdesc->seqNum); */ /* this gets handled by AMUDP_ServiceIncomingMessages */
  requestbasicbuf->status.replyIssued = TRUE;
  ep->stats.RepliesSent[category]++;
  ep->stats.DataBytesSent[category] += sizeof(int) * numargs + nbytes;
  return AM_OK;
  }

/*------------------------------------------------------------------------------------
 * Request
 *------------------------------------------------------------------------------------ */
extern int AMUDP_Request(ep_t request_endpoint, int reply_endpoint, handler_t handler, 
                         int numargs, ...) {
  AMUDP_CHECKINIT();
  if (!request_endpoint || reply_endpoint < 0) AMUDP_RETURN_ERR(BAD_ARG);
  if (handler >= AMUDP_MAX_NUMHANDLERS) AMUDP_RETURN_ERR(BAD_ARG);
  if (request_endpoint->depth == -1) AMUDP_RETURN_ERR(NOT_INIT); /* it's an error to call before AM_SetExpectedResources */
  if (reply_endpoint >= AMUDP_MAX_NUMTRANSLATIONS ||
     !request_endpoint->translation[reply_endpoint].inuse) AMUDP_RETURN_ERR(BAD_ARG);
  assert(numargs >= 0 && numargs <= AMUDP_MAX_SHORT);

  { /*  call the generic requestor */
    int retval;
    va_list argptr;
    va_start(argptr, numargs); /*  pass in last argument */
    retval = AMUDP_RequestGeneric(amudp_Short, 
                                  request_endpoint, reply_endpoint, handler, 
                                  NULL, 0, 0,
                                  numargs, argptr,
                                  amudp_system_user, 0);
    va_end(argptr);
    return retval;
    }
  }
/* ------------------------------------------------------------------------------------ */
extern int AMUDP_RequestI(ep_t request_endpoint, int reply_endpoint, handler_t handler, 
                          void *source_addr, int nbytes,
                          int numargs, ...) {
  AMUDP_CHECKINIT();
  if (!request_endpoint || reply_endpoint < 0) AMUDP_RETURN_ERR(BAD_ARG);
  if (handler >= AMUDP_MAX_NUMHANDLERS) AMUDP_RETURN_ERR(BAD_ARG);
  if (request_endpoint->depth == -1) AMUDP_RETURN_ERR(NOT_INIT); /* it's an error to call before AM_SetExpectedResources */
  if (reply_endpoint >= AMUDP_MAX_NUMTRANSLATIONS ||
     !request_endpoint->translation[reply_endpoint].inuse) AMUDP_RETURN_ERR(BAD_ARG);
  if (!source_addr) AMUDP_RETURN_ERR(BAD_ARG);
  if (nbytes < 0 || nbytes > AMUDP_MAX_MEDIUM) AMUDP_RETURN_ERR(BAD_ARG);
  assert(numargs >= 0 && numargs <= AMUDP_MAX_SHORT);

  { /*  call the generic requestor */
    int retval;
    va_list argptr;
    va_start(argptr, numargs); /*  pass in last argument */
    retval = AMUDP_RequestGeneric(amudp_Medium, 
                                  request_endpoint, reply_endpoint, handler, 
                                  source_addr, nbytes, 0,
                                  numargs, argptr,
                                  amudp_system_user, 0);
    va_end(argptr);
    return retval; 
    }
  }
/* ------------------------------------------------------------------------------------ */
static int getFreeBulkSlot(ep_t ep, int destP, uint8_t *slotnum, int allowblock) { 
  /* find a free bulk slot number, possibly blocking if allowblock=TRUE
   * returns AM_OK (with slot set), AM_ERR_XXX on error, or -1 for timeout if allowblock=FALSE
   */
  assert(ep && slotnum);
  assert(destP >= 0 && destP < ep->P);
  while(1) {
    uint8_t slot;
    for (slot = 0; slot < 16; slot++) {
      int inst;
      for (inst = 0; inst < ep->depth; inst++) {
        if (GET_REQ_DESC(ep, destP, inst)->inuse) { 
          amudp_system_messagetype_t systype =
            (amudp_system_messagetype_t)(GET_REQ_BUF(ep, destP, inst)->Msg.systemMessageType & 0xF);
          uint8_t thisslot = (uint8_t)((GET_REQ_BUF(ep, destP, inst)->Msg.systemMessageType >> 4) & 0xF);
          if (systype == amudp_system_bulkxferfragment && thisslot == slot) break; /*  already taken */
          }
        }
      if (inst == ep->depth) {
        *slotnum = slot;
        return AM_OK;
        }
      }

    /*  wait for some slots to become free */
    if (allowblock) { 
      int retval;
      retval = AMUDP_Block(ep->eb);
      if (retval != AM_OK) AMUDP_RETURN(retval);
      retval = AM_Poll(ep->eb);
      if (retval != AM_OK) AMUDP_RETURN(retval);
      }
    else return -1; /*  timed out - non-blocking */
    }
  }
/* ------------------------------------------------------------------------------------ */
extern int AMUDP_RequestXfer(ep_t request_endpoint, int reply_endpoint, handler_t handler, 
                          void *source_addr, int nbytes, uintptr_t dest_offset, 
                          int async, 
                          int numargs, ...) {
  AMUDP_CHECKINIT();
  if (!request_endpoint || reply_endpoint < 0) AMUDP_RETURN_ERR(BAD_ARG);
  if (handler >= AMUDP_MAX_NUMHANDLERS) AMUDP_RETURN_ERR(BAD_ARG);
  if (request_endpoint->depth == -1) AMUDP_RETURN_ERR(NOT_INIT); /* it's an error to call before AM_SetExpectedResources */
  if (reply_endpoint >= AMUDP_MAX_NUMTRANSLATIONS ||
     !request_endpoint->translation[reply_endpoint].inuse) AMUDP_RETURN_ERR(BAD_ARG);
  if (!source_addr) AMUDP_RETURN_ERR(BAD_ARG);
  if (nbytes < 0 || nbytes > AMUDP_MAX_LONG) AMUDP_RETURN_ERR(BAD_ARG);
  if (dest_offset > AMUDP_MAX_SEGLENGTH) AMUDP_RETURN_ERR(BAD_ARG);
  assert(numargs >= 0 && numargs <= AMUDP_MAX_SHORT);
  #if USE_TRUE_BULK_XFERS
  {
    int destP = request_endpoint->translation[reply_endpoint].id;
    if (async) { /*  decide if we can satisfy request without blocking */
      int i;
      /* it's unclear from the spec whether we should poll before an async failure,
       * but by definition the app must be prepared for handlers to run when calling this 
       * request, so it shouldn't cause anything to break, and the async request is more likely
       * to succeed if we do. so:
       */
      AM_Poll(request_endpoint->eb);

      /* see if there's a free buffer */
      for (i = 0; i < request_endpoint->depth; i++) 
        if (!request_endpoint->requestDesc[destP].inuse) break;
      if (i == request_endpoint->depth)         
        AMUDP_RETURN_ERRFR(IN_USE, AMUDP_RequestXferAsync, 
          "Request can't be satisfied without blocking right now");
      }

    { /* perform the send */
      int retval;
      va_list argptr;
      va_start(argptr, numargs); /*  pass in last argument */
      retval = AMUDP_RequestGeneric(amudp_Long, 
                                    request_endpoint, reply_endpoint, handler, 
                                    source_addr, nbytes, dest_offset,
                                    numargs, argptr,
                                    amudp_system_user, 0);
      va_end(argptr);
      return retval; 
      }
    }
  #else
  { /* segmented bulk transfers - break large xfers into chunks */
    int numchunks = ((nbytes-1) / AMUDP_MAX_MEDIUM)+1; /*  number of chunks required for data */
    int destP = request_endpoint->translation[reply_endpoint].id;
    if (async) { /*  decide if we can satisfy request without blocking */
      int i, freecnt=0;

      if (numchunks > request_endpoint->depth) 
        AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_RequestXferAsync, 
          "Request too large to EVER be satisfied without blocking");

      /* it's unclear from the spec whether we should poll before an async failure,
       * but by definition the app must be prepared for handlers to run when calling this 
       * request, so it shouldn't cause anything to break, and the async request is more likely
       * to succeed if we do. so:
       */
      AM_Poll(request_endpoint->eb);

      /*  see how many buffers are free */
      for (i = 0; i < request_endpoint->depth; i++) {
        if (!request_endpoint->requestDesc[destP].inuse) freecnt++;
        }
      if (numchunks > freecnt) 
        AMUDP_RETURN_ERRFR(IN_USE, AMUDP_RequestXferAsync, 
          "Request too large to be satisfied without blocking right now");
      { uint8_t slotnum; /*  see if we have a free slot */
        int retval = getFreeBulkSlot(request_endpoint, destP, &slotnum, FALSE);
        if (retval == -1)
          AMUDP_RETURN_ERRFR(IN_USE, AMUDP_RequestXferAsync, 
            "Request can't be satisfied without blocking right now - out of bulk xfer slots");
        }
      }
    if (numchunks == 1) { /*  single-message bulk xfer, just a user message */
      /*  call the generic requestor */
      int retval;
      va_list argptr;
      va_start(argptr, numargs); /*  pass in last argument */
      retval = AMUDP_RequestGeneric(amudp_Long, 
                                    request_endpoint, reply_endpoint, handler, 
                                    source_addr, nbytes, dest_offset,
                                    numargs, argptr,
                                    amudp_system_user, 0);
      va_end(argptr);
      return retval; 
      }
    else { 
      int fragidx;
      char *chunk_source_addr = source_addr;
      uintptr_t chunk_dest_offset = dest_offset;
      uint8_t slotnum;

      /*  get a slot (possibly poll-blocking) */
      uint8_t systype;
      int retval = getFreeBulkSlot(request_endpoint, destP, &slotnum, TRUE);
      if (retval != AM_OK) AMUDP_RETURN(retval);
      assert(slotnum < 16);
      systype = (slotnum << 4) | ((uint8_t)amudp_system_bulkxferfragment);

      for (fragidx = 0; fragidx < numchunks; fragidx++) {
        int retval;
        va_list argptr;
        uint16_t chunk_nbytes = fragidx < numchunks - 1 ? 
                                AMUDP_MAX_MEDIUM :
                                ( nbytes % AMUDP_MAX_MEDIUM == 0 ? 
                                  AMUDP_MAX_MEDIUM : 
                                  nbytes % AMUDP_MAX_MEDIUM );
        va_start(argptr, numargs); /*  pass in last argument */
        retval = AMUDP_RequestGeneric(amudp_Long, 
                                      request_endpoint, reply_endpoint, handler, 
                                      chunk_source_addr, chunk_nbytes, chunk_dest_offset,
                                      numargs, argptr,
                                      systype, (uint8_t)(numchunks-1));
        va_end(argptr);
        if (retval != AM_OK) { /*  recovery here would suck, so errors here are fatal */
          ErrMessage("Network failure in the middle of a bulk transfer");
          abort(); 
          }
        chunk_source_addr += chunk_nbytes;
        chunk_dest_offset += chunk_nbytes;
        }
      return AM_OK;
      }
    }
  #endif
  }
/*------------------------------------------------------------------------------------
 * Reply
 *------------------------------------------------------------------------------------ */
extern int AMUDP_Reply(void *token, handler_t handler, 
                       int numargs, ...) {
  amudp_buf_t *basicbuf;
  amudp_buf_t *requestbuf;

  AMUDP_CHECKINIT();
  if (!token) AMUDP_RETURN_ERR(BAD_ARG);
  if (handler >= AMUDP_MAX_NUMHANDLERS) AMUDP_RETURN_ERR(BAD_ARG);
  assert(numargs >= 0 && numargs <= AMUDP_MAX_SHORT);

  { /*  semantic checking on reply (are we in a handler, is this the first reply, etc.) */
    basicbuf = (amudp_buf_t *)token;
    requestbuf = (basicbuf->status.bulkBuffer ? basicbuf->status.bulkBuffer : basicbuf);
    if (!AMUDP_MSG_ISREQUEST(&requestbuf->Msg)) AMUDP_RETURN_ERR(RESOURCE); /* token is not a request */
    if (!basicbuf->status.handlerRunning) AMUDP_RETURN_ERR(RESOURCE); /* token is not for an active request */
    if (basicbuf->status.replyIssued) AMUDP_RETURN_ERR(RESOURCE);     /* already issued a reply */
    if (((amudp_system_messagetype_t)requestbuf->Msg.systemMessageType) != amudp_system_user) 
      AMUDP_RETURN_ERR(RESOURCE); /* can't reply to a system message (returned message) */
    }

  { /*  call the generic replier */
    int retval;
    va_list argptr;
    va_start(argptr, numargs); /*  pass in last argument */
    retval = AMUDP_ReplyGeneric(amudp_Short, 
                                  basicbuf, handler, 
                                  NULL, 0, 0,
                                  numargs, argptr,
                                  amudp_system_user, 0);
    va_end(argptr);
    return retval; 
    }
  }
/* ------------------------------------------------------------------------------------ */
extern int AMUDP_ReplyI(void *token, handler_t handler, 
                          void *source_addr, int nbytes,
                          int numargs, ...) {
  amudp_buf_t *basicbuf;
  amudp_buf_t *requestbuf;

  AMUDP_CHECKINIT();
  if (!token) AMUDP_RETURN_ERR(BAD_ARG);
  if (handler >= AMUDP_MAX_NUMHANDLERS) AMUDP_RETURN_ERR(BAD_ARG);
  if (!source_addr) AMUDP_RETURN_ERR(BAD_ARG);
  if (nbytes < 0 || nbytes > AMUDP_MAX_MEDIUM) AMUDP_RETURN_ERR(BAD_ARG);
  assert(numargs >= 0 && numargs <= AMUDP_MAX_SHORT);

  { /*  semantic checking on reply (are we in a handler, is this the first reply, etc.) */
    basicbuf = (amudp_buf_t *)token;
    requestbuf = (basicbuf->status.bulkBuffer ? basicbuf->status.bulkBuffer : basicbuf);
    if (!AMUDP_MSG_ISREQUEST(&requestbuf->Msg)) AMUDP_RETURN_ERR(RESOURCE); /* token is not a request */
    if (!basicbuf->status.handlerRunning) AMUDP_RETURN_ERR(RESOURCE); /* token is not for an active request */
    if (basicbuf->status.replyIssued) AMUDP_RETURN_ERR(RESOURCE);     /* already issued a reply */
    if (((amudp_system_messagetype_t)requestbuf->Msg.systemMessageType) != amudp_system_user) 
      AMUDP_RETURN_ERR(RESOURCE); /* can't reply to a system message (returned message) */
    }

  { /*  call the generic replier */
    int retval;
    va_list argptr;
    va_start(argptr, numargs); /*  pass in last argument */
    retval = AMUDP_ReplyGeneric(amudp_Medium, 
                                  basicbuf, handler, 
                                  source_addr, nbytes, 0,
                                  numargs, argptr,
                                  amudp_system_user, 0);
    va_end(argptr);
    return retval; 
    }
  }
/* ------------------------------------------------------------------------------------ */
extern int AMUDP_ReplyXfer(void *token, handler_t handler, 
                          void *source_addr, int nbytes, uintptr_t dest_offset, 
                          int numargs, ...) {
  amudp_buf_t *basicbuf;
  amudp_buf_t *requestbuf;

  AMUDP_CHECKINIT();
  if (!token) AMUDP_RETURN_ERR(BAD_ARG);
  if (handler >= AMUDP_MAX_NUMHANDLERS) AMUDP_RETURN_ERR(BAD_ARG);
  if (!source_addr) AMUDP_RETURN_ERR(BAD_ARG);
  if (nbytes < 0 || nbytes > AMUDP_MAX_LONG) AMUDP_RETURN_ERR(BAD_ARG);
  if (dest_offset > AMUDP_MAX_SEGLENGTH) AMUDP_RETURN_ERR(BAD_ARG);
  assert(numargs >= 0 && numargs <= AMUDP_MAX_SHORT);

  if (nbytes > AMUDP_MAX_MEDIUM && !USE_TRUE_BULK_XFERS) 
    AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_ReplyXfer, 
      "ReplyXfer() exceeding AM_Medium()=" _STRINGIFY(AMUDP_MAX_MEDIUM) " bytes not implemented.");

  { /*  semantic checking on reply (are we in a handler, is this the first reply, etc.) */
    basicbuf = (amudp_buf_t *)token;
    requestbuf = (basicbuf->status.bulkBuffer ? basicbuf->status.bulkBuffer : basicbuf);
    if (!AMUDP_MSG_ISREQUEST(&requestbuf->Msg)) AMUDP_RETURN_ERR(RESOURCE); /* token is not a request */
    if (!basicbuf->status.handlerRunning) AMUDP_RETURN_ERR(RESOURCE); /* token is not for an active request */
    if (basicbuf->status.replyIssued) AMUDP_RETURN_ERR(RESOURCE);     /* already issued a reply */
    if (((amudp_system_messagetype_t)requestbuf->Msg.systemMessageType) != amudp_system_user) 
      AMUDP_RETURN_ERR(RESOURCE); /* can't reply to a system message (returned message) */
    }


  { /*  call the generic replier */
    int retval;
    va_list argptr;
    va_start(argptr, numargs); /*  pass in last argument */
    retval = AMUDP_ReplyGeneric(amudp_Long, 
                                  basicbuf, handler, 
                                  source_addr, nbytes, dest_offset,
                                  numargs, argptr,
                                  amudp_system_user, 0);
    va_end(argptr);
    return retval; 
    }
  }
/* ------------------------------------------------------------------------------------ */
extern void AMUDP_DefaultReturnedMsg_Handler(int status, op_t opcode, void *token) {
  char *statusStr = "*unknown*";
  char *opcodeStr = "*unknown*";
  amudp_buf_t *msgbasicbuf = (amudp_buf_t *)token;
  amudp_buf_t *msgbuf = (msgbasicbuf->status.bulkBuffer ? msgbasicbuf->status.bulkBuffer : msgbasicbuf);
  int numArgs = AMUDP_MSG_NUMARGS(&msgbuf->Msg);
  uint32_t *args = GET_PACKET_ARGS(msgbuf);
  char argStr[255];
  int i;

  #define STATCASE(name, desc) case name: statusStr = #name ": " desc; break;
  switch (status) {
    STATCASE(EBADARGS      , "Arguments to request or reply function invalid    ");
    STATCASE(EBADENTRY     , "X-lation table index selected unbound table entry ");
    STATCASE(EBADTAG       , "Sender's tag did not match the receiver's EP tag  "); 
    STATCASE(EBADHANDLER   , "Invalid index into the recv.'s handler table      "); 
    STATCASE(EBADSEGOFF    , "Offset into the dest-memory VM segment invalid    ");
    STATCASE(EBADLENGTH    , "Bulk xfer length goes beyond a segment's end      ");
    STATCASE(EBADENDPOINT  , "Destination endpoint does not exist               ");
    STATCASE(ECONGESTION   , "Congestion at destination endpoint                ");
    STATCASE(EUNREACHABLE  , "Destination endpoint unreachable                  ");
    STATCASE(EREPLYREJECTED, "Destination endpoint refused reply message        ");
    }
  #define OPCASE(name) case name: opcodeStr = #name; break;
  switch (opcode) {
    OPCASE(AM_REQUEST_M);
    OPCASE(AM_REQUEST_IM);
    OPCASE(AM_REQUEST_XFER_M);
    OPCASE(AM_REPLY_M);
    OPCASE(AM_REPLY_IM);
    OPCASE(AM_REPLY_XFER_M);
    }

  argStr[0] = '\0';
  for (i=0; i < numArgs; i++) {
    char tmp[20];
    sprintf(tmp, "0x%08x  ", args[i]);
    strcat(argStr, tmp);
    }
  { char temp1[80];
    char temp2[80];
  ErrMessage("An active message was returned to sender,\n"
             "    and trapped by the default returned message handler (handler 0):\n"
             "Error Code: %s\n"
             "Message type: %s\n"
             "Destination: %s (%i)\n"
             "Handler: %i\n"
             "Tag: %s\n"
             "Arguments(%i): %s\n"
             "Aborting...",
             statusStr, opcodeStr, 
             AMUDP_enStr(msgbasicbuf->status.sourceAddr, temp1), msgbasicbuf->status.sourceId,
             msgbuf->Msg.handlerId, AMUDP_tagStr(msgbuf->Msg.tag, temp2),
             numArgs, argStr);
    }
  abort();
  }
/* ------------------------------------------------------------------------------------ */
