/*  $Archive:: /Ti/AMUDP/amudp_spmd.cpp                                   $
 *     $Date: Tue, 11 Dec 2001 16:31:30 -0800 $
 * $Revision: 1.6 $
 * Description: AMUDP Implementations of SPMD operations (bootstrapping and parallel job control)
 * Copyright 2000, Dan Bonachea <bonachea@cs.berkeley.edu>
 */

#include <assert.h>
#include <stdio.h>
#ifdef WIN32
  #define sched_yield() Sleep(0)
  #define sleep(x) Sleep(x*1000)
  #include <process.h>
#else
  #include <unistd.h>
  #include <sched.h>
  #if defined(LINUX) && !defined(__USE_GNU)
    /* some Linuxes need this to pull in F_SETSIG */
    #define __USE_GNU
    #include <fcntl.h>
    #undef __USE_GNU
  #else
    #include <fcntl.h>
  #endif
#endif

extern char **environ; 

#include <amudp.h>
#include <amudp_spmd.h>
#include <amudp_internal.h>
#include "sockutil.h"
#include "socklist.h"
#include "sig.h"

#define FD_STDIN 0
#define FD_STDOUT 1
#define FD_STDERR 2

#define FREEZE_SLAVE  0
static volatile bool amudp_frozen = true;
// all this to make sure we get a full stack frame for debugger
static void freezeForDebugger(int depth=0) {
  if (!depth) freezeForDebugger(1);
  else {
    volatile int i;
    while (amudp_frozen) {
      i++;
      sleep(1);
      }
    }
  }

#ifdef DEBUG
  #define DEBUG_SLAVE(msg)  do { fprintf(stderr,"slave %i: %s\n", AMUDP_SPMDMYPROC, msg); fflush(stderr); } while(0)
  #define DEBUG_MASTER(msg) do { fprintf(stderr,"master: %s\n", msg); fflush(stderr); } while(0)
#else
  #define DEBUG_SLAVE(msg)  do {} while(0) /* prevent silly warnings about empty statements */
  #define DEBUG_MASTER(msg) do {} while(0)
#endif

#define AMUDP_SPMDSLAVE_FLAG "__AMUDP_SLAVE_PROCESS__"


/* master only */
  static SOCKET AMUDP_SPMDListenSocket = INVALID_SOCKET; /* TCP bootstrapping listener */
  static SOCKET AMUDP_SPMDStdinListenSocket = INVALID_SOCKET; 
  static SOCKET AMUDP_SPMDStdoutListenSocket = INVALID_SOCKET; 
  static SOCKET AMUDP_SPMDStderrListenSocket = INVALID_SOCKET; 
  static SOCKET *AMUDP_SPMDSlaveSocket = NULL; /* table of TCP control sockets */
  static amudp_translation_t *AMUDP_SPMDTranslationTable = NULL;
  int AMUDP_SPMDSpawnRunning = FALSE; /* true while spawn is active */
  #if DISABLE_STDSOCKET_REDIRECT
    int AMUDP_SPMDRedirectStdsockets = FALSE; /* true if stdin/stdout/stderr should be redirected */
  #else
    int AMUDP_SPMDRedirectStdsockets = TRUE; /* true if stdin/stdout/stderr should be redirected */
  #endif

/* slave only */
  SOCKET AMUDP_SPMDControlSocket = INVALID_SOCKET; 
  static ep_t AMUDP_SPMDEndpoint = NULL;
  static eb_t AMUDP_SPMDBundle = NULL;
  static en_t AMUDP_SPMDName = {0};
  volatile int AMUDP_SPMDIsActiveControlSocket = 0; 
  static SOCKET newstdin = INVALID_SOCKET;
  static SOCKET newstdout = INVALID_SOCKET;
  static SOCKET newstderr = INVALID_SOCKET;
  static int AMUDP_SPMDMYPROC = -1;
  static volatile int AMUDP_SPMDBarrierDone = 0; /* flag barrier as complete */
  int AMUDP_SPMDwakeupOnControlActivity = 0;

/* master & slave */
  static int AMUDP_SPMDStartupCalled = 0;
  static int AMUDP_SPMDNUMPROCS = -1;
  static char *AMUDP_SPMDMasterEnvironment = NULL;

// used to pass info
typedef struct {
  int32_t procid;       // id for this processor
  int32_t numprocs;     // num procs in job
  uint64_t networkpid;  // globally unique pid
  int32_t depth;        // network depth
  tag_t tag;            // tag for this processor
  double faultInjectionRate; // AMUDP_FaultInjectionRate
  uint16_t stdinMaster; // address of stdin listener
  uint16_t stdoutMaster; // address of stdout listener
  uint16_t stderrMaster; // address of stderr listener
  uint32_t translationtablesz; // size of translation table we're about to send
  uint32_t environtablesz; // size of environment table we're about to send
  } AMUDP_SPMDBootstrapInfo_t;

/*
  Protocol for TCP bootstrapping/control sockets
  initialization: 
    slave->master (en_t) - send my endpoint name for init
    master->slave (AMUDP_SPMDBootstrapInfo_t) 
    master->slave (AMUDP_SPMDTranslationTable (variable size)) 
    master->slave (AMUDP_SPMDMasterEnvironment (variable size)) 

  master->slave messages
    "E"(int exitcode) - die now with this exit code
    "F"(old en_t)(new en_t) - some slave's NIC just failed over to new en_t
    "B" - barrier complete

  slave->master messages
    "E"(int exitcode) - exit with this code
    "F"(old en_t)(new en_t) - this slave's NIC just failed over to new en_t
    "B" - enter barrier
*/
/* ------------------------------------------------------------------------------------ 
 *  misc helpers
 * ------------------------------------------------------------------------------------ */
static void flushStreams(char *context) {
  if (!context) context = "flushStreams()";

  if (fflush(stdout)) {
    ErrMessage("failed to flush stdout in %s", context); 
    perror("fflush");
    exit(1);
    }
  if (fflush(stderr)) {
    ErrMessage("failed to flush stderr in %s", context); 
    perror("fflush");
    exit(1);
    }
  sched_yield();
  }
//------------------------------------------------------------------------------------
extern char *AMUDP_enStr(en_t en, char *buf) {
  assert(buf);
  #ifdef UFXP
    sprintf(buf, "(fixed: %i variable: %i)", en.fixed, en.variable.index);
  #else
    SockAddr tmp((sockaddr*)&en);
    sprintf(buf, "(%s:%i)", tmp.IPStr(), tmp.port());
  #endif
  return buf;
  }
extern char *AMUDP_tagStr(tag_t tag, char *buf) {
  assert(buf);
  sprintf(buf, "0x%08x%08x", 
    (uint32_t)(tag >> 32), 
    (uint32_t)(tag & 0xFFFFFFFF));
  return buf;
  }
extern const char *sockErrDesc() {
  return errorCodeString(getSocketErrorCode());
  }
//------------------------------------------------------------------------------------
static void setupStdSocket(SOCKET& ls, SocketList& list, SocketList& allList) {
  if ((int)list.getCount() < AMUDP_SPMDNUMPROCS) {
    SockAddr remoteAddr;
    SOCKET newsock = accept_socket(ls, remoteAddr);
    list.insert(newsock);
    allList.insert(newsock);
    if ((int)list.getCount() == AMUDP_SPMDNUMPROCS) {
      // close listener
      close_socket(ls);
      allList.remove(ls);
      ls = INVALID_SOCKET;
      }
    }
  else ErrMessage("master detected some unrecognized activity on a std listener");
  }
//------------------------------------------------------------------------------------
static void handleStdOutput(FILE *fd, fd_set *psockset, SocketList& list, SocketList& allList, int nproc) {
  int numset;
  static SOCKET *tempSockArr = NULL;
  if (!tempSockArr) tempSockArr = new SOCKET[nproc];
  if ((numset = list.getIntersection(psockset, tempSockArr, nproc))) { // we have some active std sockets
    for (int i=0; i < numset; i++) {
      SOCKET s = tempSockArr[i];
      assert(FD_ISSET(s, psockset));
      if (isClosed(s)) {
        DEBUG_MASTER("dropping a std output socket...");
        list.remove(s);
        allList.remove(s);
        continue;
        }
      // TODO: line-by-line buffering
      int sz = numBytesWaiting(s);
      char *buf = new char[sz+2];
      recvAll(s, buf, sz);
      buf[sz] = '\0';
      #ifdef DEBUG
        fprintf(fd, "got some output: %s", buf);
      #else
        fwrite(buf, sz, 1, fd);
      #endif
      fflush(fd);
      delete buf;
      }
    }
  }
//------------------------------------------------------------------------------------
#if USE_ASYNC_TCP_CONTROL
  static void AMUDP_SPMDControlSocketCallback(int sig) {
    AMUDP_SPMDIsActiveControlSocket = TRUE;
    #if DEBUG_VERBOSE
      fprintf(stderr, "got an AMUDP_SIGIO signal\n");fflush(stderr);
    #endif
    reghandler(AMUDP_SIGIO, AMUDP_SPMDControlSocketCallback);
    }
#endif
/* ------------------------------------------------------------------------------------ 
 *  basic inquiries
 * ------------------------------------------------------------------------------------ */
extern int AMUDP_SPMDNumProcs() {
  if (!AMUDP_SPMDStartupCalled) {
    ErrMessage("called AMUDP_SPMDNumProcs before AMUDP_SPMDStartup()");
    return -1;
    }
  assert(AMUDP_SPMDNUMPROCS >= 1);
  return AMUDP_SPMDNUMPROCS;
  }
/* ------------------------------------------------------------------------------------ */
extern int AMUDP_SPMDMyProc() {
  if (!AMUDP_SPMDStartupCalled) {
    ErrMessage("called AMUDP_SPMDMyProc before AMUDP_SPMDStartup()");
    return -1;
    }
  assert(AMUDP_SPMDMYPROC >= 0);
  return AMUDP_SPMDMYPROC;
  }
/* ------------------------------------------------------------------------------------ */
extern int AMUDP_SPMDIsWorker(char **argv) {
  if (AMUDP_SPMDStartupCalled) return 1; 
  else {
    assert(argv);
    return (argv[1] && !strcmp(argv[1], AMUDP_SPMDSLAVE_FLAG));
    }
  }
/* ------------------------------------------------------------------------------------ */
extern int AMUDP_SPMDStartup(int *argc, char ***argv,
                             int nproc, int networkdepth, 
                             spawnfn_t spawnfn,
                             uint64_t *networkpid,
                             eb_t *eb, ep_t *ep) {
  
  if (AMUDP_SPMDStartupCalled) AMUDP_RETURN_ERR(RESOURCE);
  if (!argc || !argv) AMUDP_RETURN_ERR(BAD_ARG);
  /* we need a separate socklibinit for master 
     and to prevent AM_Terminate from murdering all our control sockets */
  if (!socklibinit()) AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_SPMDStartup, "socklibinit() failed");

  /* ------------------------------------------------------------------------------------ 
   *  I'm a master 
   * ------------------------------------------------------------------------------------ */
  if ((*argc) < 2 || strcmp((*argv)[1], AMUDP_SPMDSLAVE_FLAG)) { 
    int usingdefaultdegree = 0;
    if (nproc < 0 || nproc > AMUDP_MAX_SPMDPROCS) AMUDP_RETURN_ERR(BAD_ARG);

    /* defaulting */
    if (networkdepth < 0) AMUDP_RETURN_ERR(BAD_ARG);
    if (networkdepth == 0) networkdepth = AMUDP_DEFAULT_NETWORKDEPTH;

    if (nproc == 0) { /* default to read from args */
      if (*argc > 1) nproc = atoi((*argv)[1]);
      if (nproc < 1) {
        fprintf(stderr, 
          "AMUDP SPMD Runtime Layer v%s, Copyright 2001, Dan Bonachea\n"
          "This program requires you specify the parallel degree\n"
          "as the first argument to %s\n" 
          , AMUDP_LIBRARY_VERSION_STR, (*argv)[0]);
        exit(1);
        AMUDP_RETURN_ERR(BAD_ARG);
        }
      
      usingdefaultdegree = 1;

      /* readjust params */
      (*argv)[1] = (*argv)[0];
      (*argv)++;
      (*argc)--;
      }
    AMUDP_SPMDNUMPROCS = nproc;

    if (!spawnfn) {
      int OK = 1;
      if (*argc > 1 && strlen((*argv)[1]) == 1) {
        switch ((*argv)[1][0]) {
          case 'g': case 'G':
            spawnfn = AMUDP_SPMDGlunixSpawn;
            break;
          case 's': case 'S':
            spawnfn = AMUDP_SPMDSshSpawn;
            break;
          case 'r': case 'R':
            spawnfn = AMUDP_SPMDRexecSpawn;
            break;
          case 'l': case 'L':
            spawnfn = AMUDP_SPMDLocalSpawn;
            break;
          case 'c': case 'C':
            spawnfn = AMUDP_SPMDCustomSpawn;
            break;
          default:
            OK = 0;
          }
        }
      else OK = 0;
      if (!OK) {
        fprintf(stderr, 
          "AMUDP SPMD Runtime Layer v%s, Copyright 2001, Dan Bonachea\n"
          "Usage: %s%s <spawnfn> program args...\n"
          " <spawnfn> = one of the following mechanisms for spawning remote workers:\n"
          "    'S'  Spawn jobs using ssh remote shells\n"
          "    'L'  Spawn jobs using fork()/exec() on the local machine (good for SMP's)\n"
        #ifdef REXEC
          "    'R'  Spawn jobs using rexec (Berkeley Millennium only)\n"
        #endif
        #ifdef GLUNIX
          "    'G'  Spawn jobs using GLUNIX (Berkeley NOW only)\n"
        #endif
          "    'C'  Spawn jobs using custom spawn command\n"
          , AMUDP_LIBRARY_VERSION_STR, (*argv)[0], (usingdefaultdegree?" <paralleldegree>":""));
        exit(1);
        AMUDP_RETURN_ERR(BAD_ARG);
        }

      /* readjust params */
      (*argv)[1] = (*argv)[0];
      (*argv)++;
      (*argc)--;
      }

    // setup bootstrap info 
    AMUDP_SPMDBootstrapInfo_t bootstrapinfo;
    bootstrapinfo.numprocs = nproc;
    bootstrapinfo.translationtablesz = nproc * sizeof(amudp_translation_t);
    bootstrapinfo.depth = networkdepth;
    {
      int masterpid = getpid();
      uint32_t masterIP;
      try {
        SockAddr dnsAddr = DNSLookup(getMyHostName());
        masterIP = dnsAddr.IP();
        }
      catch (xBase &exn) {
        AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_SPMDStartup, exn.why());
        }
      bootstrapinfo.networkpid = ((uint64_t)masterIP) << 32 | 
                                 (((uint64_t)masterpid) & 0xFFFF);
      if (networkpid) *networkpid = bootstrapinfo.networkpid;
      }

    { char *faultRate = getenv("AMUDP_FAULT_RATE");
      if (faultRate && atof(faultRate) != 0.0) {      
        bootstrapinfo.faultInjectionRate = atof(faultRate);
        }
      else bootstrapinfo.faultInjectionRate = 0.0;
      }

    const char *masterHostname = getMyHostName();
    #ifdef DEBUG
      printf("master host name: %s\n", masterHostname);
    #endif

    // TCP socket lists
    SocketList allList(nproc*4+10); // a list of all active sockets
    SocketList coordList(nproc);    // a list of all coordination sockets
    SocketList stdinList(nproc);    // a list of all stdin routing sockets
    SocketList stdoutList(nproc);   // a list of all stdout routing sockets
    SocketList stderrList(nproc);   // a list of all stderr routing sockets
    AMUDP_SPMDSlaveSocket = (SOCKET*)malloc(nproc * sizeof(SOCKET));

    try {
      // create our TCP listen ports 
      unsigned short anyport = 0;
      AMUDP_SPMDListenSocket = listen_socket(anyport, false);
      AMUDP_SPMDStdinListenSocket = listen_socket(anyport, false);
      AMUDP_SPMDStdoutListenSocket = listen_socket(anyport, false);
      AMUDP_SPMDStderrListenSocket = listen_socket(anyport, false);
      bootstrapinfo.stdinMaster = getsockname(AMUDP_SPMDStdinListenSocket).port();
      bootstrapinfo.stdoutMaster = getsockname(AMUDP_SPMDStdoutListenSocket).port();
      bootstrapinfo.stderrMaster = getsockname(AMUDP_SPMDStderrListenSocket).port();
      }
    catch (xBase &exn) {
      AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_SPMDStartup, exn.why());
      }

    allList.insert(AMUDP_SPMDListenSocket);
    allList.insert(AMUDP_SPMDStdinListenSocket);
    allList.insert(AMUDP_SPMDStdoutListenSocket);
    allList.insert(AMUDP_SPMDStderrListenSocket);

    // create and initialize the translation table that we'll fill in as slaves connect
    AMUDP_SPMDTranslationTable = (amudp_translation_t*)malloc(bootstrapinfo.translationtablesz);
    for (int i=0; i < nproc; i++) {
      AMUDP_SPMDSlaveSocket[i] = INVALID_SOCKET;
      AMUDP_SPMDTranslationTable[i].tag = bootstrapinfo.networkpid | ((uint64_t)i) << 16;
      }

    {
      /* flatten a snapshot of the master's environment for transmission to slaves
       * here we assume the standard representation where a pointer to the environment 
       * is stored in a global variable 'environ' and the environment is represented as an array 
       * of null-terminated strings where each has the form 'key=value' and value may be empty, 
       * and the final string pointer is a NULL pointer
       * we flatten this into a list of null-terminated 'key=value' strings, 
       * terminated with a double-null
       */
      int i;
      int totalEnvSize = 0;
      for(i = 0; environ[i]; i++) 
        totalEnvSize += strlen(environ[i]) + 1;
      totalEnvSize++;

      AMUDP_SPMDMasterEnvironment = (char *)malloc(totalEnvSize);
      char *p = AMUDP_SPMDMasterEnvironment;
      p[0] = '\0';
      for(i = 0; environ[i]; i++) {
        strcpy(p, environ[i]);
        p += strlen(p) + 1;
        }
      *p = '\0';
      assert((p+1) - AMUDP_SPMDMasterEnvironment == totalEnvSize);
      bootstrapinfo.environtablesz = totalEnvSize;
      }

    // setup a slave argv
    char **slaveargv = (char**)malloc(sizeof(char*)*((*argc)+3));
    int slaveargc = (*argc)+2;
    slaveargv[0] = (*argv)[0];
    slaveargv[1] = AMUDP_SPMDSLAVE_FLAG;
    SockAddr masterAddr = getsockname(AMUDP_SPMDListenSocket);
    #if USE_NUMERIC_MASTER_ADDR
      slaveargv[2] = masterAddr.FTPStr();
    #else
      char masteraddrstr[1024];
      sprintf(masteraddrstr, "%s:%i", masterHostname, masterAddr.port());
      slaveargv[2] = masteraddrstr;
    #endif
    for (int k = 1; k < (*argc); k++) {
      slaveargv[k+2] = (*argv)[k];
      }
    slaveargv[slaveargc] = NULL;

    // call system-specific spawning routine
    AMUDP_SPMDSpawnRunning = TRUE;
    if (!spawnfn(nproc, slaveargc, slaveargv)) { 
      ErrMessage("Error spawning SPMD worker threads. Exiting...");
      exit(1);
      }
    AMUDP_SPMDSpawnRunning = FALSE;

    if (!AMUDP_SPMDRedirectStdsockets) {
      // spawn function disabled our stdsocket redirect - signal the slaves of this fact
      bootstrapinfo.stdinMaster = 0;
      bootstrapinfo.stdoutMaster = 0;
      bootstrapinfo.stderrMaster = 0;
      }

    // main communication loop for master
    try {
      int numSlavesAttached = 0;

      fd_set sockset;
      fd_set* psockset = &sockset;
      int numset; // helpers for coord socket
      SOCKET *tempSockArr = new SOCKET[nproc];
      while (1) {
        allList.makeFD_SET(psockset);

        if (select(allList.getMaxFd()+1, psockset, NULL, NULL, NULL) == -1) { // block for activity
          perror("select");
          exit(1);
          }
        //------------------------------------------------------------------------------------
        // stdin/stderr/stdout listeners - incoming connections
        if (AMUDP_SPMDStdinListenSocket != INVALID_SOCKET &&
            FD_ISSET(AMUDP_SPMDStdinListenSocket, psockset))  
          setupStdSocket(AMUDP_SPMDStdinListenSocket, stdinList, allList);
        if (AMUDP_SPMDStdoutListenSocket != INVALID_SOCKET &&
            FD_ISSET(AMUDP_SPMDStdoutListenSocket, psockset)) 
          setupStdSocket(AMUDP_SPMDStdoutListenSocket, stdoutList, allList);
        if (AMUDP_SPMDStderrListenSocket != INVALID_SOCKET &&
            FD_ISSET(AMUDP_SPMDStderrListenSocket, psockset)) 
          setupStdSocket(AMUDP_SPMDStderrListenSocket, stderrList, allList);
        //------------------------------------------------------------------------------------
        // stdout/err sockets - must come before possible exit to drain output
        handleStdOutput(stdout, psockset, stdoutList, allList, nproc);
        handleStdOutput(stderr, psockset, stderrList, allList, nproc);
        // stdin (illegal to receive anything here)
        if ((numset = stdinList.getIntersection(psockset, tempSockArr, nproc))) {
          for (int i=0; i < numset; i++) {
            SOCKET s = tempSockArr[i];
            assert(FD_ISSET(s, psockset));
            if (isClosed(s)) {
              DEBUG_MASTER("dropping a stdinList socket...");
              stdinList.remove(s);
              allList.remove(s);
              }
            else {
              ErrMessage("Master got illegal input on a stdin socket");
              stdinList.remove(s); // prevent subsequent warnings
              allList.remove(s);
              }
            }
          }
        //------------------------------------------------------------------------------------
        // coordination listener
        if (AMUDP_SPMDListenSocket != INVALID_SOCKET && 
            FD_ISSET(AMUDP_SPMDListenSocket, psockset)) { // incoming connection on coordination socket
          //DEBUG_MASTER("got some activity on AMUDP_SPMDListenSocket");
          if (numSlavesAttached < nproc) { // attach a slave
            SockAddr remoteAddr;
            SOCKET newcoord = accept_socket(AMUDP_SPMDListenSocket, remoteAddr);
            coordList.insert(newcoord);
            allList.insert(newcoord);

            #if USE_COORD_KEEPALIVE
            { // make sure we get connection termination notification in a timely manner
              int val = 1;
              if (setsockopt(newcoord, SOL_SOCKET, SO_KEEPALIVE, (char *)&val, sizeof(int)) == SOCKET_ERROR)
                DEBUG_MASTER("failed to setsockopt(SO_KEEPALIVE) on coord socket");
              }
            #endif

            // receive bootstrapping info
            AMUDP_SPMDSlaveSocket[numSlavesAttached] = newcoord;
            recvAll(newcoord, &(AMUDP_SPMDTranslationTable[numSlavesAttached].name), sizeof(en_t));

            numSlavesAttached++;
            if (numSlavesAttached == nproc) { // all have now reported in, so we can begin computation
              // close listener
              close_socket(AMUDP_SPMDListenSocket);
              allList.remove(AMUDP_SPMDListenSocket);
              AMUDP_SPMDListenSocket = INVALID_SOCKET;

              // transmit bootstrapping info
              for (int i=0; i < nproc; i++) {
                // fill out process-specific bootstrap info
                bootstrapinfo.procid = i;
                bootstrapinfo.tag = AMUDP_SPMDTranslationTable[i].tag;
                // send it
                sendAll(AMUDP_SPMDSlaveSocket[i], &bootstrapinfo, sizeof(bootstrapinfo));
                sendAll(AMUDP_SPMDSlaveSocket[i], AMUDP_SPMDTranslationTable, bootstrapinfo.translationtablesz);
                sendAll(AMUDP_SPMDSlaveSocket[i], AMUDP_SPMDMasterEnvironment, bootstrapinfo.environtablesz);
                }
              #ifdef DEBUG
                printf("Endpoint table (nproc=%i):\n", nproc);
                for (int j=0; j < nproc; j++) {
                  char temp[80];
                  printf(" P#%i:\t%s", j, AMUDP_enStr(AMUDP_SPMDTranslationTable[j].name, temp));
                  printf("\ttag: %s\n", AMUDP_tagStr(AMUDP_SPMDTranslationTable[j].tag, temp));
                  }
              #endif
              }
            }
          else ErrMessage("master detected some unrecognized activity on AMUDP_SPMDListenSocket");
          }
        //------------------------------------------------------------------------------------
        // coord sockets
        if ((numset = coordList.getIntersection(psockset, tempSockArr, nproc))) { // we have some active coord sockets
          //DEBUG_MASTER("got some activity on coord sockets");
          for (int i=0; i < numset; i++) {
            SOCKET s = tempSockArr[i];
            assert(FD_ISSET(s, psockset));
            if (isClosed(s)) {
              DEBUG_MASTER("dropping a coordList socket...\n");
              coordList.remove(s);
              allList.remove(s);

              #if ABORT_JOB_ON_NODE_FAILURE
                int exitCode = -1;
                for (int i=0; i < (int)coordList.getCount(); i++) {
                  sendAll(coordList[i], "E");
                  sendAll(coordList[i], &exitCode, sizeof(int));
                  close_socket(coordList[i]);
                  }
                if (!socklibend()) ErrMessage("master failed to socklibend()");
                DEBUG_MASTER("Lost a worker process - job aborting...");
                exit(exitCode);
              #endif
              continue;
              }
            char command;
            recvAll(s, &command, 1);
            switch(command) {
              case 'B': { // enter barrier
                static int AMUDP_SPMDBarrierCount = 0; /* number of processors that have entered barrier */
                AMUDP_SPMDBarrierCount++;
                if (AMUDP_SPMDBarrierCount == nproc) { // barrier complete
                  DEBUG_MASTER("Completed barrier");
                  // broadcast completion message
                  for (int i=0; i < (int)coordList.getCount(); i++) {
                    sendAll(coordList[i], "B");
                    }
                  AMUDP_SPMDBarrierCount = 0;
                  }
                break;
                }

            #ifdef UFXP
              case 'F': { // NIC fail-over
                // get relevant en_t's
                en_t olden;
                en_t newen;
                try {
                  recvAll(s, &olden, sizeof(en_t));
                  recvAll(s, &newen, sizeof(en_t));
                  }
                catch (xSocket& exn) {
                  ErrMessage("got exn while reading fail-over addresses: %s", exn.why());
                  }
                // update our local table 
                int j;
                for (j = 0; j < nproc; j++) {
                  if (enEqual(AMUDP_SPMDTranslationTable[j].name, olden)) {
                    AMUDP_SPMDTranslationTable[j].name = newen;
                    break;
                    }
                  }
                if (j == nproc) ErrMessage("unrecognized endpoint received in fail-over message");
                // tell all other slaves about the change
                for (int i=0; i < (int)coordList.getCount(); i++) {
                  sendAll(coordList[i], "F");
                  sendAll(coordList[i], &olden, sizeof(en_t));
                  sendAll(coordList[i], &newen, sizeof(en_t));
                  }
                #if DEBUG
                  char temp[80];
                  printf("master: processed NIC failover on slave %i: ", j);
                  printf("%s ->", AMUDP_enStr(olden, temp));
                  printf(" %s\n", AMUDP_enStr(newen, temp));
                #endif
                break;
                }
            #endif

              case 'E': { // exit code
                // get slave terminate code
                int exitCode = -1;
                try {
                  recvAll(s, &exitCode, sizeof(int));
                  }
                catch (xSocket& exn) {
                  ErrMessage("got exn while reading exit code: %s", exn.why());
                  }
                // tell all other slaves to terminate
                // TODO: perhaps use an active message for this? for now, just rely on coord socket dying
                // TODO: it's possblie we can lose some final output because coord and std are asynchronous
                for (int i=0; i < (int)coordList.getCount(); i++) {
                  sendAll(coordList[i], "E");
                  sendAll(coordList[i], &exitCode, sizeof(int));
                  close_socket(coordList[i]);
                  }
                if (!socklibend()) ErrMessage("master failed to socklibend()");
                #ifdef DEBUG
                  printf("Exiting after AMUDP_SPMDExit(%i)...\n", exitCode);
                #endif
                exit(exitCode);
                break;
                }

              default:
                ErrMessage("master got an unknown command on coord socket: %c", command);
              }
            }
          if (coordList.getCount() == 0) {
            DEBUG_MASTER("Exiting after losing all worker slave connections (noone called AMUDP_Exit())\n");
            exit(0); // program exit, noone called terminate
            }
          }
        //------------------------------------------------------------------------------------
        } // loop
      }
    catch (xSocket& exn) {
      ErrMessage("Master got an xSocket: %s", exn.why());
      exit(1);
      }
    catch (xBase& exn) {
      ErrMessage("Master got an xBase: %s", exn.why());
      exit(1);
      }
    }
  /* ------------------------------------------------------------------------------------ 
   *  I'm a worker slave 
   * ------------------------------------------------------------------------------------ */
  else {  
    int temp;

    #if FREEZE_SLAVE
      freezeForDebugger();
    #endif

    if (!eb || !ep) AMUDP_RETURN_ERR(BAD_ARG);
    if (AM_Init() != AM_OK) {
      ErrMessage("Failed to AM_Init() in AMUDP_SPMDStartup");
      AMUDP_RETURN_ERRFR(RESOURCE, AMUDP_SPMDStartup, "AM_Init() failed");
      }

    // parse special args 
    SockAddr masterAddr;
    if ((*argc) < 3) ErrMessage("Missing arguments to slave process");
    {
      #if USE_NUMERIC_MASTER_ADDR
        masterAddr = SockAddr((*argv)[2]);
      #else
        char *IPStr = new char[strlen((*argv)[2])+10];
        strcpy(IPStr, (*argv)[2]);
        char *portStr = strchr(IPStr, ':');
        if (!portStr) {
          ErrMessage("Malformed address argument passed to slave:'%s' (missing port)", (*argv)[2]);
          AMUDP_RETURN_ERR(BAD_ARG);
          }
        int masterPort = atoi(portStr+1);
        if (masterPort < 1 || masterPort > 65535) {
          ErrMessage("Malformed address argument passed to slave:'%s' (bad port=%i)", (*argv)[2], masterPort);
          AMUDP_RETURN_ERR(BAD_ARG);
          }
        (*portStr) = '\0';
        try {
          masterAddr = SockAddr((uint32_t)DNSLookup(IPStr).IP(), (uint16_t)masterPort);
          }
        catch (xSocket &exn) {
          AMUDP_RETURN_ERRFR(RESOURCE, "slave failed DNSLookup on master host name", exn.why());
          }
        delete IPStr;
      #endif
      (*argv)[2] = (*argv)[0]; // strip off our special args
      (*argv) += 2;
      (*argc) -= 2;
      }

    try {
      #ifdef DEBUG
        fprintf(stderr, "slave connecting to %s:%i\n", masterAddr.IPStr(), masterAddr.port());
        fflush(stderr);
      #endif

      AMUDP_SPMDControlSocket = connect_socket(masterAddr);

      #if USE_COORD_KEEPALIVE
      { // make sure we get connection termination notification in a timely manner
        int val = 1;
        if (setsockopt(AMUDP_SPMDControlSocket, SOL_SOCKET, SO_KEEPALIVE, (char *)&val, sizeof(int)) == SOCKET_ERROR)
          DEBUG_MASTER("failed to setsockopt(SO_KEEPALIVE) on coord socket");
        }
      #endif

      #ifndef UFXP
        /* here we assume the interface used to contact the master is the same 
           one to be used for UDP endpoints */
        SockAddr myinterface = getsockname(AMUDP_SPMDControlSocket);
        AMUDP_SetUDPInterface(myinterface.IP());
      #endif
        
      /* create endpoint and get name */
      temp = AM_AllocateBundle(AM_SEQ, &AMUDP_SPMDBundle);
      if (temp != AM_OK) {
        ErrMessage("Failed to create bundle in AMUDP_SPMDStartup");
        AMUDP_RETURN(temp);
        }
      temp = AM_AllocateEndpoint(AMUDP_SPMDBundle, &AMUDP_SPMDEndpoint, &AMUDP_SPMDName);
      if (temp != AM_OK) {
        ErrMessage("Failed to create endpoint in AMUDP_SPMDStartup");
        AMUDP_RETURN(temp);
        }

      // send our endpoint name to the master
      sendAll(AMUDP_SPMDControlSocket, &AMUDP_SPMDName, sizeof(AMUDP_SPMDName));

      // get information from master 
      // get the bootstrap info and translation table
      AMUDP_SPMDBootstrapInfo_t bootstrapinfo;
      recvAll(AMUDP_SPMDControlSocket, &bootstrapinfo, sizeof(AMUDP_SPMDBootstrapInfo_t));
      
      // unpack the bootstrapping info
      AMUDP_SPMDNUMPROCS = bootstrapinfo.numprocs;
      AMUDP_SPMDMYPROC = bootstrapinfo.procid;
      if (networkpid) *networkpid = bootstrapinfo.networkpid;

      // sanity checking on bootstrap info
      assert(AMUDP_SPMDNUMPROCS > 0);
      assert(AMUDP_SPMDMYPROC >= 0 && AMUDP_SPMDMYPROC < AMUDP_SPMDNUMPROCS);
      assert(bootstrapinfo.translationtablesz == bootstrapinfo.numprocs * sizeof(amudp_translation_t)); 
      assert(AMUDP_SPMDMYPROC >= 0 && AMUDP_SPMDMYPROC < AMUDP_SPMDNUMPROCS);

      amudp_translation_t *tempTranslationTable = (amudp_translation_t *)malloc(bootstrapinfo.translationtablesz);
      assert(tempTranslationTable);
      recvAll(AMUDP_SPMDControlSocket, tempTranslationTable, bootstrapinfo.translationtablesz);

      assert(tempTranslationTable[AMUDP_SPMDMYPROC].tag == bootstrapinfo.tag);
      assert(enEqual(tempTranslationTable[AMUDP_SPMDMYPROC].name, AMUDP_SPMDName));

      // setup translation table
      for (int i = 0; i < bootstrapinfo.numprocs; i++) {
        temp = AM_Map(AMUDP_SPMDEndpoint, i, tempTranslationTable[i].name, tempTranslationTable[i].tag);
        if (temp != AM_OK) {
          ErrMessage("Failed to AM_Map() in AMUDP_SPMDStartup");
          AMUDP_RETURN(temp);
          }
        }

      free(tempTranslationTable);
      tempTranslationTable = NULL;

      // receive snapshot of master environment
      AMUDP_SPMDMasterEnvironment = (char *)malloc(bootstrapinfo.environtablesz);
      assert(AMUDP_SPMDMasterEnvironment);
      recvAll(AMUDP_SPMDControlSocket, AMUDP_SPMDMasterEnvironment, bootstrapinfo.environtablesz);
      
      /* allocate network buffers */
      temp = AM_SetExpectedResources(AMUDP_SPMDEndpoint, AMUDP_SPMDNUMPROCS, bootstrapinfo.depth);
      if (temp != AM_OK) {
        ErrMessage("Failed to AM_SetExpectedResources() in AMUDP_SPMDStartup");
        AMUDP_RETURN(temp);
        }
      
      // set tag
      temp = AM_SetTag(AMUDP_SPMDEndpoint, bootstrapinfo.tag);
      if (temp != AM_OK) {
        ErrMessage("Failed to AM_SetTag() in AMUDP_SPMDStartup");
        AMUDP_RETURN(temp);
        }

      #if !DISABLE_STDSOCKET_REDIRECT
        if (bootstrapinfo.stdinMaster) {
            // perform stdin/out/err redirection
            newstdin  = connect_socket(SockAddr(masterAddr.IP(),bootstrapinfo.stdinMaster));
            newstdout = connect_socket(SockAddr(masterAddr.IP(),bootstrapinfo.stdoutMaster));
            newstderr = connect_socket(SockAddr(masterAddr.IP(),bootstrapinfo.stderrMaster));
            #if 0
              // disable buffering
              setvbuf(stdin, NULL, _IONBF, 0);
              setvbuf(stdout, NULL, _IONBF, 0);
              setvbuf(stderr, NULL, _IONBF, 0);
            #endif
            #ifdef WIN32
              #if 0
              // not sure how to do this on Win32 yet - maybe use _fdopen() and/or _fileno()
              {FILE* newf;
               if( ( newf = fopen( "c:\\data", "w" ) ) == NULL )
               {
                  puts( "Can't open file 'data'\n" );
                  exit( 1 );
               }
               if( ( newf = _fdopen( newstdout, "w" ) ) == NULL )
               {
                  puts( "fdopen failed\n" );
                  exit( 1 );
               }
              if (dup2(_fileno(newf), FD_STDOUT) < 0) { // redirect stdout to socket
                perror("dup2(stdout)");
                _exit(1); 
                }
              printf("yomama\n");
              fflush(stdout);
              fclose(newf);
              exit(0);
                }
              #endif
            #else
              /* UNIX */
              if (dup2(newstdin, FD_STDIN) < 0) { // redirect stdout to socket
                perror("dup2(stdin)");
                _exit(1); 
                }
              if (dup2(newstdout, FD_STDOUT) < 0) { // redirect stdout to socket
                perror("dup2(stdout)");
                _exit(1); 
                }
              if (dup2(newstderr, FD_STDERR) < 0) { // redirect stdout to socket
                perror("dup2(stderr)");
                _exit(1); 
                }
            #endif
          }
     #endif

      if (bootstrapinfo.faultInjectionRate != 0.0) {
        AMUDP_FaultInjectionEnabled = 1;
        AMUDP_FaultInjectionRate = bootstrapinfo.faultInjectionRate;
        fprintf(stderr, "*** Warning: AMUDP running with fault injection enabled. Rate = %6.2f %%\n",
          100.0 * AMUDP_FaultInjectionRate);
        fflush(stderr);
        }
      }
    catch (xSocket& exn) {
      ErrMessage("Got an xSocket while spawning slave process: %s", exn.why());
      exit(1);
      }

    *eb = AMUDP_SPMDBundle;
    *ep = AMUDP_SPMDEndpoint;
    AMUDP_SPMDStartupCalled = 1;

    #if USE_ASYNC_TCP_CONTROL
      // enable async notification
      reghandler(AMUDP_SIGIO, AMUDP_SPMDControlSocketCallback);
      if (fcntl(AMUDP_SPMDControlSocket, F_SETOWN, getpid())) {
        perror("fcntl(F_SETOWN, getpid())");
        ErrMessage("Failed to fcntl(F_SETOWN, getpid()) on TCP control socket - try disabling USE_ASYNC_TCP_CONTROL");
        abort();
        }
      if (fcntl(AMUDP_SPMDControlSocket, F_SETSIG, AMUDP_SIGIO)) {
        ErrMessage("Failed to fcntl(F_SETSIG, AMUDP_SIGIO) on TCP control socket - try disabling USE_ASYNC_TCP_CONTROL");
        perror("fcntl(F_SETSIG)");
        abort();
        }
      if (fcntl(AMUDP_SPMDControlSocket, F_SETFL, O_ASYNC|O_NONBLOCK)) { 
        perror("fcntl(F_SETFL, O_ASYNC|O_NONBLOCK)");
        ErrMessage("Failed to fcntl(F_SETFL, O_ASYNC|O_NONBLOCK) on TCP control socket - try disabling USE_ASYNC_TCP_CONTROL");
        abort();
        }
    #endif

    #ifdef DEBUG
    { char temp[80];
      tag_t tag;
      AM_GetTag(AMUDP_SPMDEndpoint, &tag);
      fprintf(stderr, "Slave %i/%i starting (tag=%s)...\n", 
        AMUDP_SPMDMyProc(), AMUDP_SPMDNumProcs(), AMUDP_tagStr(tag, temp));
      fflush(stderr);
      }
    #endif

    return AM_OK;
    }
  /* ------------------------------------------------------------------------------------ */
  abort(); /* never reach here */
  return AM_OK;
  }

/* ------------------------------------------------------------------------------------ 
 *  worker control handler
 * ------------------------------------------------------------------------------------ */
// called by slave to handle traffic on control socket
// sets controlMessagesServiced to indicate how many message serviced
extern int AMUDP_SPMDHandleControlTraffic(int *controlMessagesServiced) {
  if (AMUDP_SPMDControlSocket == INVALID_SOCKET) return AM_OK; // not running in SPMD mode
  #ifdef USE_ASYNC_TCP_CONTROL
    if (!AMUDP_SPMDIsActiveControlSocket) return AM_OK; // nothing to do
    ASYNC_TCP_DISABLE();
    AMUDP_SPMDIsActiveControlSocket = FALSE; 
  #endif 
  if (controlMessagesServiced) *controlMessagesServiced = 0;
  
  while (1) { // service everything waiting
    try {
      if (!inputWaiting(AMUDP_SPMDControlSocket)) {
        ASYNC_TCP_ENABLE();
        return AM_OK; // nothing more to do
        }
      } 
    catch (xBase &exn) {
      ErrMessage("Error checking AMUDP_SPMDControlSocket: %s", exn.why()); // probably conn reset
      }

    try {
      SOCKET s = AMUDP_SPMDControlSocket;

      if (isClosed(s)) {
        DEBUG_SLAVE("master control socket slammed shut. Exiting...\n");
        exit(1);
        }

      // there's something waiting on the control socket for us - grab it
      char command;
      recvAll(s, &command, 1);
      switch(command) {
        case 'B': { // barrier complete
          assert(!AMUDP_SPMDBarrierDone);
          AMUDP_SPMDBarrierDone = 1; // flag completion
          break;
          }

      #ifdef UFXP
        case 'F': { // NIC fail-over
          // get relevant en_t's
          en_t olden;
          en_t newen;
          try {
            recvAll(s, &olden, sizeof(en_t));
            recvAll(s, &newen, sizeof(en_t));
            }
          catch (xSocket& exn) {
            ErrMessage("got exn while reading fail-over addresses: %s", exn.why());
            exit(1);
            }

          // this update could be rather slow, but we expect it to run extremely infrequently

          // update all translation tables
          for (int i = 0; i < AMUDP_SPMDBundle->n_endpoints; i++) {
            ep_t ep = AMUDP_SPMDBundle->endpoints[i];
            assert(ep);

            for (int j = 0; j < AMUDP_MAX_NUMTRANSLATIONS; j++) {
              if (ep->translation[j].inuse) {
                if (enEqual(ep->translation[j].name, olden)) { // need to re-map
                  ep->translation[j].name = newen;
                  }
                }
              }
            for (int k = 0; k < ep->P; k++) {
              if (enEqual(ep->perProcInfo[k].remoteName, olden)) { // need to re-map
                ep->perProcInfo[k].remoteName = newen;
                }
              }

            #ifndef UFXP
              // update any messages already accepted into the rx buffers from this guy
              // this currently never runs because UFXP has no explicit receive buffers we can see
              for (int i = 0; i < AMUDP_SPMDBundle->n_endpoints; i++) {
                ep_t ep = AMUDP_SPMDBundle->endpoints[i];
                assert(ep);

                for (int j = ep->rxReadyIdx; j != ep->rxFreeIdx; j = (j+1)%ep->rxNumBufs) {
                  if (enEqual(ep->rxBuf[j].source, olden) {
                    ep->rxBuf[j].source = newen;
                    }
                  }
                }
            #endif
            }
          #if DEBUG
            char temp[80];
            printf("slave: handled NIC failover: ");
            printf("%s ->", AMUDP_enStr(olden, temp));
            printf(" %s\n", AMUDP_enStr(newen, temp));
          #endif
          break;
          }
      #endif

        case 'E': { // exit code
          // get slave terminate code
          int exitCode = -1;
          try {
            recvAll(s, &exitCode, sizeof(int));
            }
          catch (xSocket& exn) {
            ErrMessage("got exn while reading exit code: %s", exn.why());
            }
          #ifdef DEBUG
            printf("Exiting after exit signal from master (%i)...\n", exitCode);
          #endif
            // important that we make this call to free resources
          if (AM_Terminate() != AM_OK) ErrMessage("Failure in AM_Terminate()");
          exit(exitCode);
          break;
          }

        default:
          ErrMessage("slave got an unknown command on coord socket: %c", command);
          exit(1); // this is a fatal error
        }
      }
    catch (xSocket& exn) {
      ErrMessage("Slave got an xSocket: %s. Exiting...", exn.why());
      exit(1);
      }
    catch (xBase& exn) {
      ErrMessage("Slave got an xBase: %s. Exiting...", exn.why());
      exit(1);
      }
    if (controlMessagesServiced) (*controlMessagesServiced)++;
    }
  }
/* ------------------------------------------------------------------------------------ 
 *  handler for NIC fail-over
 * ------------------------------------------------------------------------------------ */
#ifdef UFXP
extern void AMUDP_SPMDAddressChangeCallback(ufxp_addr_t *address) {
  assert(AMUDP_UFXP_endpoint);
  assert(address);
  en_t olden = AMUDP_UFXP_endpoint->name;
  en_t newen = *address;

  AMUDP_UFXP_endpoint->name = newen;
  AMUDP_SPMDName = newen;

  // send change to master, who will propagate new address info to peers and back to us
  // we update our translation table on recieving the reflection from the master
  try {
    sendAll(AMUDP_SPMDControlSocket, "F");
    sendAll(AMUDP_SPMDControlSocket, &olden, sizeof(en_t));
    sendAll(AMUDP_SPMDControlSocket, &newen, sizeof(en_t));
    }
  catch (xSocket& exn) {
    ErrMessage("Slave got an xSocket: %s. Exiting...", exn.why());
    exit(1);
    }
  catch (xBase& exn) {
    ErrMessage("Slave got an xBase: %s. Exiting...", exn.why());
    exit(1);
    }

  #ifndef UFXP
    // update any messages already accepted into the rx buffers
    // this currently never runs because UFXP has no explicit receive buffers we can see
    for (int i = 0; i < AMUDP_SPMDBundle->n_endpoints; i++) {
      ep_t ep = AMUDP_SPMDBundle->endpoints[i];
      assert(ep);

      for (int j = ep->rxReadyIdx; j != ep->rxFreeIdx; j = (j+1)%ep->rxNumBufs) {
        if (enEqual(ep->rxBuf[j].dest, olden) {
          ep->rxBuf[j].dest = newen;
          }
        }
      }
  #endif
  }
#endif
/* ------------------------------------------------------------------------------------ 
 *  process termination
 * ------------------------------------------------------------------------------------ */
extern int AMUDP_SPMDExit(int exitcode) {
  if (!AMUDP_SPMDStartupCalled) AMUDP_RETURN_ERR(NOT_INIT);
  if (AM_Terminate() != AM_OK) 
    ErrMessage("failed to AM_Terminate() in AMUDP_SPMDExit()");

  flushStreams("AMUDP_SPMDExit");

  if (fclose(stdin)) {
    ErrMessage("failed to fclose stdout in AMUDP_SPMDExit()"); 
    perror("fclose");
    }
  if (fclose(stdout)) {
    ErrMessage("failed to fclose stdout in AMUDP_SPMDExit()"); 
    perror("fclose");
    }
  if (fclose(stderr)) {
    ErrMessage("failed to fclose stderr in AMUDP_SPMDExit()"); 
    perror("fclose");
    }

  /* use normal shutdown and closesocket to ignore errors */
  if (newstdin != INVALID_SOCKET) {
    shutdown(newstdin, SHUT_RDWR);
    closesocket(newstdin); 
    }
  if (newstdout != INVALID_SOCKET) {
    shutdown(newstdout, SHUT_RDWR);
    closesocket(newstdout); 
    }
  if (newstderr != INVALID_SOCKET) {
    shutdown(newstderr, SHUT_RDWR);
    closesocket(newstderr);
    }

  sched_yield();

  /* notify master we're exiting */
  try {
    ASYNC_TCP_DISABLE();
    sendAll(AMUDP_SPMDControlSocket, "E");
    sendAll(AMUDP_SPMDControlSocket, &exitcode, sizeof(int));
    while (1) { // swallow everything and wait for master to close
      char temp;
      int retval = recv(AMUDP_SPMDControlSocket, &temp, 1, 0); 
      if (retval == 0 || retval == SOCKET_ERROR) break;
      }
    }
  catch (xBase& ) { } // ignore errors that may happen on conn reset 
  closesocket(AMUDP_SPMDControlSocket);

  if (!socklibend()) ErrMessage("slave failed to socklibend()");

  AMUDP_SPMDStartupCalled = 0;
  DEBUG_SLAVE("exiting..");
  exit(exitcode);
  abort();
  return AM_OK;
  }
/* ------------------------------------------------------------------------------------ 
 *  barrier
 * ------------------------------------------------------------------------------------ */
extern int AMUDP_SPMDBarrier() {
  if (!AMUDP_SPMDStartupCalled) {
    ErrMessage("called AMUDP_SPMDBarrier before AMUDP_SPMDStartup()");
    AMUDP_RETURN_ERR(NOT_INIT);
    }

  flushStreams("AMUDP_SPMDBarrier");
  assert(AMUDP_SPMDBarrierDone == 0);
  ASYNC_TCP_DISABLE();
  sendAll(AMUDP_SPMDControlSocket, "B");
  ASYNC_TCP_ENABLE();
  #if USE_BLOCKING_SPMD_BARRIER
    {
      int oldmask;
      AM_GetEventMask(AMUDP_SPMDBundle, &oldmask);
      // wait for completion
      AM_Poll(AMUDP_SPMDBundle);
      while (!AMUDP_SPMDBarrierDone) {
        AM_SetEventMask(AMUDP_SPMDBundle, AM_NOTEMPTY);
        AMUDP_SPMDwakeupOnControlActivity = 1;
        AM_WaitSema(AMUDP_SPMDBundle);
        AMUDP_SPMDwakeupOnControlActivity = 0;
        AM_Poll(AMUDP_SPMDBundle);
        }
      AM_SetEventMask(AMUDP_SPMDBundle, oldmask);
      }
  #else
  {
    uint32_t timeoutusec = 100;
    AM_Poll(AMUDP_SPMDBundle);
    while (!AMUDP_SPMDBarrierDone) {

      struct timeval tv;
      tv.tv_sec  = timeoutusec / 1000000;
      tv.tv_usec = timeoutusec % 1000000;
      select(1, NULL, NULL, NULL, &tv); /* sleep a little while */

      AM_Poll(AMUDP_SPMDBundle);
      if (timeoutusec < 10000) timeoutusec *= 2;
      }
    }
  #endif
  AMUDP_SPMDBarrierDone = 0;
  DEBUG_SLAVE("Leaving barrier");
  return AM_OK;
  }
/* ------------------------------------------------------------------------------------ 
 *  global getenv()
 * ------------------------------------------------------------------------------------ */
extern const char* AMUDP_SPMDgetenvMaster(const char *keyname) {
  if (!AMUDP_SPMDStartupCalled) {
    ErrMessage("called AMUDP_SPMDgetenvMaster before AMUDP_SPMDStartup()");
    return NULL;
    }

  assert(AMUDP_SPMDMasterEnvironment);
  char *p = AMUDP_SPMDMasterEnvironment;
  if (!keyname) return NULL;
  int keylen = strlen(keyname);
  while (*p) {
    if (!strncmp(keyname, p, keylen) && p[keylen] == '=') {
      return p + keylen + 1;
      }
    p += strlen(p) + 1;
    }
  return NULL; // not found
  }
/* ------------------------------------------------------------------------------------ */

