/*  $Archive:: /Ti/AMUDP/ufxp.h                                           $
 *     $Date: Wed, 26 Sep 2001 01:45:46 -0700 $
 * $Revision: 1.1 $
 * Description: UFXP interface to ISTORE-specific hardware driver
 * Copyright 2000, Daniel Hettena <danielh@cs.berkeley.edu>
 *                 and Dan Bonachea <bonachea@cs.berkeley.edu>
 */

#ifndef _UFXP_H
#define _UFXP_H

#ifdef WIN32
/* ufxp doesn't support WIN32, this is just to allow static checking of win amudp apps */
#include <stddef.h>
typedef unsigned __int8		u8;
typedef unsigned __int16	u16;
typedef unsigned __int32	u32;
typedef unsigned __int64	u64;
typedef __int8			s8;
typedef __int16			s16;
typedef __int32			s32;
typedef __int64			s64;
#else
typedef unsigned char		u8;
typedef unsigned short		u16;
typedef unsigned int		u32;
typedef unsigned long long	u64;
typedef signed char		s8;
typedef signed short        	s16;
typedef signed int          	s32;
typedef signed long long    	s64;
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  short fixed;
  struct {
    short index;
#ifdef HWADDR    
    u8 eth[6];
#endif    
  } variable;
} ufxp_addr_t;

/* the length of this must be a multiple of UFXP_ALIGNMENT */  
typedef struct {
  u8	dst[6];
  u8	src[6];
  u16	proto;
  u16	checksum;
#ifdef SWLENGTH  
  u16	len;
#endif  
#ifdef HWADDR  
  u16	srcnode;
  u16	srcindex;
#endif  
} ufxp_header_t;

#define UFXP_HEADERLENGTH	sizeof(ufxp_header_t)	/* subject to change */
#define UFXP_MAXPACKETSIZE	1514
#define UFXP_ALIGNMENT		8			/* subject to change */

typedef	void	(*ufxp_addr_hook_t)(ufxp_addr_t *address);

/* return values for functions */
#define UFXP_OK                 0   /* operation completed successfully */
#define UFXP_ERR_TIMEDOUT       1   /* a blocking operation timed out */
#define UFXP_ERR_MEMRESOURCE    2   /* insuf. physical memory or invalid alignment */
#define UFXP_ERR_NICRESOURCE    3   /* no working network device found */
#define UFXP_ERR_NOTINIT        4   /* ufxp layer has not been initialized */

/*  works without init */
s64   ufxp_getcputime();	/* strictly increasing w/ time */
s64   ufxp_getustime();		/* (SLOW) time in microseconds since who-knows-when */

int   ufxp_init();
/* ensures card can be allocated, and ufxp_getaddress can subsequently be called 
 * returns UFXP_OK or UFXP_ERR_NICRESOURCE 
 */

int   ufxp_allocatepool(void **sendpool, unsigned int maxpacketsize,
			int maxsends, int maxrecvs);
/* Must be called exactly once after ufxp_init().
 * No packets longer than maxpacketsize will be sendable or receivable.
 *
 * Guarantees (sendpool % UFXP_ALIGNMENT) == 0.
 * The send pool will be exactly maxpacketsize*maxsends bytes long.
 *
 * Returns: UFXP_OK, UFXP_ERR_NOTINIT, UFXP_ERR_MEMRESOURCE
 */

int  ufxp_getaddress(ufxp_addr_t *address);
int  ufxp_setaddresshook(ufxp_addr_hook_t hook);
/* both work any time after init
 * hook will only be called during send/recv 
 * returns UFXP_OK, UFXP_ERR_NOTINIT
 */

/* first call init, setrecvpool, setaddresshook; send/recv assert this */

int   ufxp_set_packet_destination(void *msg, ufxp_addr_t *dst);
int   ufxp_send_preset(const void *msg, unsigned int len);

static inline int ufxp_send(void *msg, unsigned int len, ufxp_addr_t *dst) {
  int result = ufxp_set_packet_destination(msg, dst);
  if (result != UFXP_OK)
    return result;

  return ufxp_send_preset(msg, len);
}
/* if send fails (i.e. out of ethernet cards), crashes. 
 * returns UFXP_OK, UFXP_ERR_NOTINIT, UFXP_ERR_NICRESOURCE
 * the message passed in includes space for the ufxp header 
 * the actual data payload starts at msg[UFXP_HEADERLENGTH]
 * (and the len argument includes this extra space) 
 * loopback messages (those destined for this NIC) will be handled correctly
 */

int   ufxp_recv(void **buf, unsigned int *len, int blockus, ufxp_addr_t *src);
/* must call ufxp_freerxbuf(buf) later to release the receive buffer
 * blockus:
 * -1	block indefinitely
 * 0	strict poll
 * n>0	wait n microseconds until time out
 * recv does ping/pong to make sure eth is alive 
 * returns UFXP_OK, UFXP_ERR_NOTINIT, UFXP_ERR_NICRESOURCE, UFXP_ERR_TIMEDOUT
 * the returned buffer points to the ufxp header and returned len includes this header
 *	space; also, len is always at least 60, even if the packet is smaller.
 * the actual data payload starts at buf[UFXP_HEADERLENGTH]
 * the packet and the actual data payload are guaranteed to be aligned by UFXP_ALIGNMENT
 * if len is NULL, the parameter is ignored and the client gets 
 *  no indication about the size of the recv'd message
 * if src is NULL, the parameter is ignored and the client gets 
 *  no indication of the sender's address
 * if buf is NULL, the function behaves the same with respect to blocking behavior
 *  (blocking/polling/waiting as indicated until the recieve pool becomes non-empty)
 *  but no message, len or source address is returned and the recieve pool remains
 *  unchanged by the call
 * if ufxp_recv returns UFXP_OK when buf is NULL, the next call to ufxp_recv is
 *  guaranteed to return a message immediately without blocking
 */

int  ufxp_freerxbuf(void *buf);
/* release a receive buffer that was returned by ufxp_recv
 * returns UFXP_OK, UFXP_ERR_NOTINIT, UFXP_ERR_MEMRESOURCE
 */


int  ufxp_terminate();
/* terminate operations and release all resources (NIC and any allocated memory)
 * returns UFXP_OK, UFXP_ERR_NOTINIT
 * if the application terminates or crashes without calling this function, 
 * the resources will still be properly freed
 */

/* General notes:
 * no reentrancy:
 *  the layer will work in a multi-threaded application, 
 *  but none of the functions may be called concurrently
 *  (so accesses must be controlled by locks or other logic to ensure this)
 */

#ifdef __cplusplus
}
#endif


#endif
