#ifndef __monitor_c
#define __monitor_c

#include <tic.h>
#include <sys/time.h>
#include <unistd.h>
#include <gc.h>
#include <native-utils.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#define TRUE 1
#define FALSE 0

#define DEBUG_MONITORS 0

#ifdef DEBUG_MONITORS
#define ASSERT(x) assert(x)
#else
#define ASSERT(x)
#endif


void monitor_startup() {
	COMM_GetMyProcess()->locked_monitors      = NULL;	
	COMM_GetMyProcess()->signal_proc          = (Process)-1;	
	COMM_GetMyProcess()->signal_box           = (Process)-1;	
	COMM_GetMyProcess()->waiting_for_monitor  = 0;	
}

void monitor_destroy(tic_monitor_t *monitor) {

  waiting_proc *temp;

  if (monitor != NULL) {
    /* Hopefully, the monitor will never be deallocated before */
    /* replies to all processes waiting to enter the monitor are sent. */
    /* However, this is done just in case this situation arises. */
    while (monitor->waitingProcs != NULL) {
      temp = monitor->waitingProcs;
      monitor->waitingProcs = monitor->waitingProcs->next;
      temp->next = NULL;
      ti_free(temp);
    }
    __tic_lock_destroy(&(monitor->lock));
    __tic_cond_destroy(&(monitor->condvar));
  }

}


void monitor_enter(jGPointer monitor) {
#ifndef MEMORY_SHARED
  	return;
#else

  tic_monitor_t *mon = (tic_monitor_t *)TO_LOCAL(monitor);

  	/* Do I already own it? */
 if (mon->proc == MYPROC) {
    	mon->nesting++;
    	return;
  }

  __tic_lock(&(mon->lock));
  ASSERT(mon->proc == (Process)-1);
  ASSERT(mon->nesting == 0);
  mon->proc = MYPROC;
  mon->nesting = 1;
  COMM_thread_cons_monitor(monitor);
#endif /* NO_SMP */

}

void monitor_exit(jGPointer monitor) {
#ifndef MEMORY_SHARED
    return;
#else

  tic_monitor_t *mon = (tic_monitor_t *)TO_LOCAL(monitor);

    	/* Do I own it? */
    if (mon->proc == MYPROC) {
      	if (mon->nesting > 1) {
		mon->nesting--;
		return;
      	} else {
        	ASSERT(mon->nesting == 1);
		mon->nesting = 0;
		mon->proc = (Process)-1;
		COMM_thread_remove_monitor(monitor);
		__tic_unlock(&(mon->lock));
		return;
      	}
    } else { /* I can't NOT own it, by the matching semantics of monitors */
      	ASSERT(0);
      	abort();
    }
#endif /* NO_SMP */
}

void monitor_wait(jGPointer monitor, ___tic_time_t *tm) {
#ifndef MEMORY_SHARED
  return;
#else
  int oldnesting;
  tic_monitor_t *mon = (tic_monitor_t *)TO_LOCAL(monitor);

   	/* Do I own it? */
  if (mon->proc == MYPROC) {
      		/* Therefore, I own the condvarlock too. */
      	ASSERT(mon->nesting > 0);
      	oldnesting = mon->nesting;
      	mon->nesting = 0;
      	mon->proc = (Process)-1;
      	COMM_thread_remove_monitor(monitor);
      	if (tm == NULL) __tic_mon_wait(mon); 
      	else            __tic_mon_timedwait(mon, tm); 
      	ASSERT(mon->proc == (Process)-1);
      	ASSERT(mon->nesting == 0);
      	COMM_thread_cons_monitor(monitor);
      	mon->proc = MYPROC;
      	mon->nesting = oldnesting;
   } else {
  	tossMonitorStateException_str("Tried to wait on a lock you didn't own.\n");
  	abort();			
   }
#endif /* NO_SMP */

}

void notify_waiter(tic_monitor_t *mon) {
    
#ifndef MEMORY_SHARED
    return;
#else
    ASSERT(mon->proc == MYPROC);
    ASSERT(mon->nesting > 0);
    __tic_mon_signal(mon);
#endif /* NO_SMP */

}

void notify_waiters(tic_monitor_t *mon) {
    
#ifndef MEMORY_SHARED 
    return;
#else 
    ASSERT(mon->proc == MYPROC);
    ASSERT(mon->nesting > 0);
    __tic_mon_broadcast(mon);
#endif /* NO_SMP */
}

void monitor_notify(jGPointer monitor) {
#ifndef MEMORY_SHARED
    return;
#else

    tic_monitor_t *mon = (tic_monitor_t *)TO_LOCAL(monitor);

    	/* Do I already own it? */
    if (mon->proc == MYPROC) {
      	ASSERT(mon->nesting > 0);
      	notify_waiter(mon);
      	return;
    } else {
  	tossMonitorStateException_str("Tried to notify a lock you didn't own.\n");
  	abort();			
    }
#endif /* NO_SMP */

}

void monitor_notify_all(jGPointer monitor) {
#ifndef MEMORY_SHARED
    return;
#else

    tic_monitor_t *mon = (tic_monitor_t *)TO_LOCAL(monitor);

    	/* Do I already own it? */
    if (mon->box == MYBOX && mon->proc == MYPROC) {
      	ASSERT(mon->nesting > 0);
      	notify_waiters(mon);
      	return;
    } else {
  	tossMonitorStateException_str("Tried to notify a lock you didn't own.\n");
  	abort();			
    }
#endif /* NO_SMP */
}

#endif  /* __monitor_c */
