/* store.h - Global memory accesses without acks (store)
 *           Implementation dependent definitions and declarations.
 */
/* see copyright.txt for usage terms */

#ifndef __STORE_H__
#define __STORE_H__

#include <tic.h>

/* .....................
 *
 * Profiling information
 *
 * .....................
 */

extern __tic_lock_t Store_mutex;

/* .....................
 *
 * Keep track of STORESs
 *
 * .....................
 */

void store_init(void);
void all_store_sync(void);
void all_store_sync_ctr(Counter *rbytes, Counter *sbytes);

/* A local variable which is incremented and then we wait for it to go to zero.
   Used as the implicit flag in all non-blocking operations */
extern Counter recvBytes;
extern Counter sentBytes;

/* Forget about all stores */
INLINE_MODIFIER void reset_store_sync(void)
INLINE({
	__tic_lock(&Store_mutex);
	sentBytes -= recvBytes;
	recvBytes = 0;
	__tic_unlock(&Store_mutex);
})


/* Wait for all stores *to* this processor to complete */
INLINE_MODIFIER void store_sync(int expecting)
INLINE({
	while(recvBytes < expecting) {
		tic_poll();
	}
	__tic_lock(&Store_mutex);
	sentBytes -= expecting;
	recvBytes -= expecting;
	__tic_unlock(&Store_mutex);
})

INLINE_MODIFIER void store_sync_ctr(int expecting,
				  Counter *rBytes, Counter *sBytes)
INLINE({
	while(*rBytes < expecting) {
	  tic_poll();
	}
	__tic_lock(&Store_mutex);
	*sBytes -= expecting;
	*rBytes -= expecting;
	__tic_unlock(&Store_mutex);
})


/* Check whether all stores *to* this processor have completed */
INLINE_MODIFIER int is_store_sync(int expecting)
INLINE({
	tic_poll();
	return recvBytes >= expecting;
})

INLINE_MODIFIER int is_store_sync_ctr(int expecting,
					     Counter *rbytes,
					     Counter *sbytes __attribute__((unused)))
INLINE({
	tic_poll();
	return *rbytes >= expecting;
})

#ifdef DEPRECATED
INLINE_MODIFIER int store_rcvd_bytes(void)
INLINE({
	return recvBytes;
})
#endif

INLINE_MODIFIER void __incr_ctr_size(Counter *ctr, int size)
INLINE({
  __tic_lock(&Store_mutex);
  (*ctr) += size;
  __tic_unlock(&Store_mutex);
})


/* ............................
 *
 * STOREs with explicit counter
 *
 * ............................
 */

#ifdef COMM_AM2
/* Handler on the remote end to store value */
extern void __b_store_handler (void *token, jbyte *addr, jbyte data,
			       Counter *ctr);
extern void __sh_store_handler(void *token, jshort *addr, jshort data,
			       Counter *ctr);
extern void __i_store_handler (void *token, jint *addr, jint data,
			       Counter *ctr);
extern void __f_store_handler (void *token, jfloat *addr, jfloat data,
			       Counter *ctr);

extern void __d_store_handler (void *token, jdouble *addr, jdouble data,
			       Counter *ctr);
extern void __l_store_handler(void *token, jlong *addr, jlong data,
			      Counter *ctr);
extern void __lp_store_handler(void *token, void **addr, void *data,
			      Counter *ctr);
#ifdef PTR64
extern void __gp_store_handler (void *token, jGPointer *addr, 
			      jUIntPointer data0, jUIntPointer data1, 
			      Counter *ctr);
#else
extern void __gp_store_handler (void *token, jGPointer *addr, 
			      jGPointer data,
			      Counter *ctr);
#endif

#endif /* COMM_AM2 */

/* STORE character */
INLINE_MODIFIER void __b_store_ctr(jGPointer gPtr, jbyte val,
				 Counter *rbytes, Counter *sbytes)
#ifdef COMM_AM2
     INLINE({
       int box = tobox(gPtr);
       jbyte *addr    = tolocal(gPtr);

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jbyte));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jbyte));
       }
       else {
	 __incr_ctr_size(sbytes, sizeof(jbyte));
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request3(CUR_ENDPOINT, box, __B_STORE_H, (jUIntPointer)addr, 
		     val, (jUIntPointer)rbytes);
       }
     })
#else
     INLINE({
       jbyte *addr    = tolocal(gPtr);
	 __incr_ctr_size(sbytes, sizeof(jbyte));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jbyte));
     })
#endif /* COMM_AM2 */


/* STORE short */
INLINE_MODIFIER void __sh_store_ctr(jGPointer gPtr, jshort val,
				  Counter *rbytes, Counter *sbytes)
#ifdef COMM_AM2
     INLINE({
       int box = tobox(gPtr);
       jshort *addr   = tolocal(gPtr);

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jshort));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jshort));
       } 
       else {
	 __incr_ctr_size(sbytes, sizeof(jshort));
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request3(CUR_ENDPOINT, box, __SH_STORE_H, (jUIntPointer)addr, 
		     val, (jUIntPointer)rbytes);
       }
     })
#else
     INLINE({
       jshort *addr   = tolocal(gPtr);
	 __incr_ctr_size(sbytes, sizeof(jshort));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jshort));
     })
#endif /* COMM_AM2 */


/* STORE int */
INLINE_MODIFIER void __i_store_ctr(jGPointer gPtr, int val,
				 Counter *rbytes, Counter *sbytes)
#ifdef COMM_AM2
     INLINE({
       int box = tobox(gPtr);
       jint *addr     = tolocal(gPtr);

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jint));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jint));
       } 
       else {
	 __incr_ctr_size(sbytes, sizeof(jint));
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request3(CUR_ENDPOINT, box, __I_STORE_H, (jUIntPointer)addr, 
		     val, (jUIntPointer)rbytes);
       }
     })
#else
     INLINE({
       jint *addr     = tolocal(gPtr);
	 __incr_ctr_size(sbytes, sizeof(jint));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jint));
     })
#endif /* COMM_AM2 */

/* STORE float */
INLINE_MODIFIER void __f_store_ctr(jGPointer gPtr, jfloat val,
				 Counter *rbytes, Counter *sbytes)
#ifdef COMM_AM2
     INLINE({
       int box = tobox(gPtr);
       jfloat *addr   = tolocal(gPtr);

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jfloat));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jfloat));
       } 
       else {
	 __incr_ctr_size(sbytes, sizeof(jfloat));
	 /* pass unsigned so that it won't pass in fp registers */
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request3(CUR_ENDPOINT, box, __F_STORE_H, (jUIntPointer)addr, 
		     val, (jUIntPointer)rbytes);
       }
     })
#else
     INLINE({
       jfloat *addr   = tolocal(gPtr);
	 __incr_ctr_size(sbytes, sizeof(jfloat));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jfloat));
     })
#endif /* COMM_AM2 */


/* STORE double */
INLINE_MODIFIER void __d_store_ctr(jGPointer gPtr, jdouble val,
				      Counter *rbytes, Counter *sbytes)
#ifdef COMM_AM2
#ifdef PTR64
     INLINE({
       int box = tobox(gPtr);
       jdouble *addr   = tolocal(gPtr);

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jdouble));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jdouble));
       } 
       else {
	 __incr_ctr_size(sbytes, sizeof(jdouble));
	 /* pass unsigned so it will not pass in fp regs */
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request3(CUR_ENDPOINT, box, __D_STORE_H, (jUIntPointer)addr, 
			val, (jUIntPointer)rbytes);
       }
     })
#else /* PTR64 */
     INLINE({
       int box = tobox(gPtr);
       jdouble *addr   = tolocal(gPtr);

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jdouble));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jdouble));
       } 
       else {
	 __incr_ctr_size(sbytes, sizeof(jdouble));
	 /* pass unsigned so it will not pass in fp regs */
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request4(CUR_ENDPOINT, box, __D_STORE_H, (jUIntPointer)addr, 
			HIGHWORD(*((julong *)&val)), 
			LOWWORD(*((julong *)&val)), (jUIntPointer)rbytes);
       }
     })
#endif /* PTR64 */
#else
     INLINE({
       jdouble *addr   = tolocal(gPtr);

       __incr_ctr_size(sbytes, sizeof(jdouble));
       *addr = val;
       __incr_ctr_size(rbytes, sizeof(jdouble));
     })
#endif /* COMM_AM2 */     



/* STORE long long */
INLINE_MODIFIER void __l_store_ctr(jGPointer gPtr, jlong val,
				 Counter *rbytes, Counter *sbytes)
#ifdef COMM_AM2
#ifdef PTR64
     INLINE({
       int box = tobox(gPtr);
       jlong *addr   = tolocal(gPtr);

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jlong));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jlong));
       } 
       else {
	 __incr_ctr_size(sbytes, sizeof(jlong));
	 /* pass unsigned so it will not pass in fp regs */
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request3(CUR_ENDPOINT, box, __L_STORE_H, (jUIntPointer)addr, 
			val, (jUIntPointer)rbytes);
       }
     })
#else /* PTR64 */
     INLINE({
       int box = tobox(gPtr);
       jlong *addr   = tolocal(gPtr);

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jlong));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jlong));
       } 
       else {
	 __incr_ctr_size(sbytes, sizeof(jdouble));
	 /* pass unsigned so it will not pass in fp regs */
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request4(CUR_ENDPOINT, box, __L_STORE_H, (jUIntPointer)addr, 
		     HIGHWORD((julong)val), 
		     LOWWORD((julong)val), (jUIntPointer)rbytes);
       }
     })
#endif /* PTR64 */
#else
     INLINE({
       jlong *addr   = tolocal(gPtr);

       __incr_ctr_size(sbytes, sizeof(jlong));
       *addr = val;
       __incr_ctr_size(rbytes, sizeof(jlong));
     })
#endif /* COMM_AM2 */     

/* STORE local pointer */
INLINE_MODIFIER void __lp_store_ctr(jGPointer gPtr, void *val,
				 Counter *rbytes, Counter *sbytes)
#ifdef COMM_AM2
     INLINE({
       int box = tobox(gPtr);
       void **addr   = tolocal(gPtr);

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(void *));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(void *));
       } 
       else {
	 __incr_ctr_size(sbytes, sizeof(void *));
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request3(CUR_ENDPOINT, box, __LP_STORE_H, (jUIntPointer)addr, 
			(jUIntPointer)val, (jUIntPointer)rbytes);
       }
     })
#else
     INLINE({
       void **addr   = tolocal(gPtr);

       __incr_ctr_size(sbytes, sizeof(void *));
       *addr = val;
       __incr_ctr_size(rbytes, sizeof(void *));
     })
#endif /* COMM_AM2 */     


/* STORE global pointer */
INLINE_MODIFIER void __gp_store_ctr(jGPointer gPtr, jGPointer val,
				  Counter *rbytes, Counter *sbytes)
#ifdef COMM_AM2
#ifdef PTR64
     INLINE({
       int box = tobox(gPtr);
       jGPointer *addr   = tolocal(gPtr);
       jUIntPointer data0;
       jUIntPointer data1;

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jGPointer));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jGPointer));
       } 
       else {
	 data0 = HIGHWORD(val);
	 data1 = LOWWORD(val);
	 __incr_ctr_size(sbytes, sizeof(jGPointer));
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request4(CUR_ENDPOINT, box, __GP_STORE_H, (jUIntPointer)addr, 
			data0, data1, (jUIntPointer)rbytes);
       }
     })
#else /* PTR64 */
     INLINE({
       int box = tobox(gPtr);
       jGPointer *addr   = tolocal(gPtr);
       Box boxData;
       Process procData; 
       void *addrData;

       if(box == MYBOX) {
	 __incr_ctr_size(sbytes, sizeof(jGPointer));
	 *addr = val;
	 __incr_ctr_size(rbytes, sizeof(jGPointer));
       } 
       else {
	 boxData = tobox(val);
	 procData = toproc(val);
	 addrData = tolocal(val);
	 __incr_ctr_size(sbytes, sizeof(jGPointer));
	 ADVANCE_SEND_ENDPOINT;
	 TI_AM_Request5(CUR_ENDPOINT, box, __GP_STORE_H, (jUIntPointer)addr, 
			boxData, procData, (jUIntPointer)addrData, (jUIntPointer)rbytes);
       }
     })
#endif /* PTR64 */
#else
     INLINE({
       jGPointer *addr   = tolocal(gPtr);

       __incr_ctr_size(sbytes, sizeof(jGPointer));
       *addr = val;
       __incr_ctr_size(rbytes, sizeof(jGPointer));
     })
#endif /* COMM_AM2 */     





/* ............................
 *
 * STOREs with implicit counter
 *
 * ............................
 */

INLINE_MODIFIER void __b_store(jGPointer gPtr, jbyte val)
INLINE({ __b_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

INLINE_MODIFIER void __sh_store(jGPointer gPtr, jshort val)
INLINE({ __sh_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

INLINE_MODIFIER void __i_store(jGPointer gPtr, jint val)
INLINE({
	__i_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

INLINE_MODIFIER void __f_store(jGPointer gPtr, jfloat val)
INLINE({ __f_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

INLINE_MODIFIER void __d_store(jGPointer gPtr, jdouble val)
INLINE({ __d_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

INLINE_MODIFIER void __l_store(jGPointer gPtr, jlong val)
INLINE({ __l_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

INLINE_MODIFIER void __lp_store(jGPointer gPtr, void * val)
INLINE({ __lp_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

INLINE_MODIFIER void __gp_store(jGPointer gPtr, jGPointer val)
INLINE({ __gp_store_ctr (gPtr, val, &recvBytes, &sentBytes); })

#if 0 
INLINE_MODIFIER void __struct_i_store(jGPointer gPtr, ___Struct_i val)
INLINE({ __i_store_ctr (gPtr, val.data, &recvBytes, &sentBytes); })

INLINE_MODIFIER void __struct_d_store(jGPointer gPtr, ___Struct_d val)
INLINE({ __ll_store_ctr (gPtr, val.data, &recvBytes, &sentBytes); })
#endif
	
extern void __store(jGPointer gPtr, void *lPtr, int len);




#ifdef PTR64

#define ___STORE_HANDLER4_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _store_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data, \
					    juint ctr0, juint ctr1);


#define ___STORE_HANDLER4(shorttype, type) \
void ___ ## shorttype ## _store_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data, \
					    juint ctr0, juint ctr1) { \
  type *addr; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _store_handler(token, addr, (type)data, ctr); \
} 

#define ___STORE_HANDLER8_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _store_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data0, juint data1,  \
					    juint ctr0, juint ctr1);

#define ___STORE_HANDLER8(shorttype, type) \
void ___ ## shorttype ## _store_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data0, juint data1,  \
					    juint ctr0, juint ctr1) { \
  type *addr; \
  type data; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&data, data0); \
  SETLOWWORD((julong *)&data, data1); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _store_handler(token, addr, data, ctr); \
}

#define ___STORE_HANDLER16(shorttype, type, halftype) \
extern void ___ ## shorttype ## _store_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data00, juint data01,  \
					    juint data02, juint data03,  \
					    juint ctr0, juint ctr1);

#define ___STORE_HANDLER16(shorttype, type, halftype) \
void ___ ## shorttype ## _store_handler(void *token, \
					    juint addr0, juint addr1, \
					    juint data00, juint data01,  \
					    juint data02, juint data03,  \
					    juint ctr0, juint ctr1) { \
  type *addr; \
  halftype data0, data1; \
  Counter *ctr; \
  \
  SETHIGHWORD((julong *)&addr, addr0); \
  SETLOWWORD((julong *)&addr, addr1); \
  SETHIGHWORD((julong *)&data0, data00); \
  SETLOWWORD((julong *)&data0, data01); \
  SETHIGHWORD((julong *)&data1, data02); \
  SETLOWWORD((julong *)&data1, data03); \
  SETHIGHWORD((julong *)&ctr, ctr0); \
  SETLOWWORD((julong *)&ctr, ctr1); \
  \
  __ ## shorttype ## _store_handler(token, addr, data0, data1, ctr); \
}

 
#else

#define ___STORE_HANDLER4_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _store_handler(void *token, juint addr,  \
				      juint data0, juint ctr);

#define ___STORE_HANDLER4(shorttype, type) \
void ___ ## shorttype ## _store_handler(void *token, juint addr,  \
				      juint data0, juint ctr) {  \
  __ ## shorttype ## _store_handler(token, (type *)addr, (type)data0, \
				  (Counter *)ctr); \
}

#define ___STORE_HANDLER8_PTYPE(shorttype, type) \
extern void ___ ## shorttype ## _store_handler(void *token, juint addr,  \
				      juint data0, juint data1, juint ctr);

#define ___STORE_HANDLER8(shorttype, type) \
void ___ ## shorttype ## _store_handler(void *token, juint addr,  \
				      juint data0, juint data1, juint ctr) {  \
  type data; \
  SETHIGHWORD((julong *)&data, data0); \
  SETLOWWORD((julong *)&data, data1); \
  __ ## shorttype ## _store_handler(token, (type *)addr, data, \
				        (Counter *)ctr); \
}


#endif  /* PTR64 */

___STORE_HANDLER4_PTYPE(b, jbyte)
___STORE_HANDLER4_PTYPE(sh, jshort)
___STORE_HANDLER4_PTYPE(i, jint)
___STORE_HANDLER4_PTYPE(f, jfloat)
___STORE_HANDLER8_PTYPE(d, jdouble)
___STORE_HANDLER8_PTYPE(l, jlong)
#ifdef PTR64
     ___STORE_HANDLER16_PTYPE(gp, jGPointer, jUIntPointer)
     ___STORE_HANDLER8_PTYPE(lp, void *)
#else
void ___gp_store_handler(void *token, juint addr, juint box, juint proc, juint addrData,
		       juint ctr);
     ___STORE_HANDLER4_PTYPE(lp, void *)
#endif







#endif /* __STORE_H__ */
