/* Provides an interval type.  Defines the type interval with fields
lo and hi.  The user should #define the following:

    interval
    make_interval                            
    singleton_interval
    copy_interval                            
    is_empty
    include_in_interval
    INCLUDE_INTERVAL_DIVISION (optional)
    divide_interval*
    divide_rounding_to_minus_inf*
    divide_rounding_to_plus_inf*
    multiply_interval
    shift_interval                            
    containing_interval                            
    intersect_intervals
    interval_contains (optional)
    ALLOC

    Those marked * are needed only if INCLUDE_INTERVAL_DIVISION is true.
*/

#include "alloc.h"
#include "bool.h"

typedef struct {
  T lo, hi;
} interval;

INLINE_MODIFIER interval *make_interval(T lo, T hi)
INLINE ({
  interval *r = ALLOC(interval, 1);
  r->lo = lo;
  r->hi = hi;
  assert(lo <= hi);
  return r;
})

INLINE_MODIFIER interval *singleton_interval(T p)
INLINE ({
  return make_interval(p, p);
})

INLINE_MODIFIER interval *copy_interval(interval *i)
INLINE ({
  return make_interval(i->lo, i->hi);
})

INLINE_MODIFIER bool is_empty(interval *i)
INLINE ({
  return (i == NULL || i->lo > i->hi);
})

/* Modify i to contain a. */
INLINE_MODIFIER void include_in_interval(interval *i, T a)
INLINE ({
  if (is_empty(i))
    i->lo = i->hi = a;
  else {
    if (i->lo > a)
      i->lo = a;
    if (i->hi < a)
      i->hi = a;
  }
})

#ifdef interval_contains
INLINE_MODIFIER bool interval_contains(interval *i, T a)
INLINE ({
  return (i->lo <= a && a <= i->hi);
})
#endif

#if INCLUDE_INTERVAL_DIVISION
/* Return r, the smallest interval such that i is a subset of d times r. */
INLINE_MODIFIER interval *divide_interval(interval *i, T d)
INLINE ({
  if (!is_empty(i)) {
    if (d != 1) {
      i->lo = divide_rounding_to_minus_inf(i->lo, d);
      i->hi = divide_rounding_to_plus_inf(i->hi, d);
    }
    if (d < 0) {
      T x = i->lo;
      i->lo = i->hi;
      i->hi = x;
    } 
  }
  return i;
})
#endif

INLINE_MODIFIER interval *multiply_interval(interval *i, T d)
INLINE ({
  if (d != 1) {
    i->lo *= d;
    i->hi *= d;
  }
  if (d < 0) {
    T x = i->lo;
    i->lo = i->hi;
    i->hi = x;
  } 
  return i;
})

INLINE_MODIFIER interval *shift_interval(interval *i, T d)
INLINE ({
  i->lo += d;
  i->hi += d;
  return i;
})

/* Return a new interval that is the smallest interval that covers i and j. */
INLINE_MODIFIER interval *containing_interval(interval *i, interval *j)
INLINE ({
  if (is_empty(i))
    return make_interval(j->lo, j->hi);
  else if (is_empty(j))
    return make_interval(i->lo, i->hi);
  else {
    interval *r = ALLOC(interval, 1);
    r->lo = ((i->lo < j->lo) ? i->lo : j->lo);
    r->hi = ((i->hi > j->hi) ? i->hi : j->hi);
    return r;
  }
})

/* Return a new interval that is intersection of i and j. */
INLINE_MODIFIER interval *intersect_intervals(interval *i, interval *j)
INLINE ({
  if (is_empty(i))
    return make_interval(i->lo, i->hi);
  else if (is_empty(j))
    return make_interval(j->lo, j->hi);
  else {
    interval *r = ALLOC(interval, 1);
    r->lo = ((i->lo > j->lo) ? i->lo : j->lo);
    r->hi = ((i->hi < j->hi) ? i->hi : j->hi);
    return r;
  }
})

