#include "AC.h"
#include "OrderingConstraint.h"
#include "BoundingBox.h"
#include "linalg.h"
#include "storage.h"
#include <algorithm>

Megatile::~Megatile()
{
  free_list_and_delete_elements(loops);
  ::free_all(m);
}

Megatile::Megatile(Foreach *f)
{
  loops = ::cons(f);
  m = NULL;
  n = 1;
  for (int i = f->k; --i >= 0; )
    m = ::cons(0, ::cons(i, m));
  usib = NULL;
}

void Megatile::summarize_ts(ostream &o)
{
  if (!reader_to_writer.empty()) {
    o << "TS summary: " << endl;
    {
      map< intpair, st_source * >::const_iterator i;
      for (i = reader_to_writer.begin(); i != reader_to_writer.end(); ++i) {
	const intpair &read = (*i).first;
	int acc = read.second, k = read.first;
	const st_source *write = (*i).second;
	o << "  Read acc=" << acc << ",k=" << k << " comes from " <<
	  write->to_string() << endl;
      }
    }
    map< smap_at_k, string >::const_iterator i;
    for (i = writer_to_name.begin(); i != writer_to_name.end(); ++i) {
      const smap_at_k &writer = (*i).first;
      const string &name = (*i).second;
      const intvector *v = writer_to_maxdist[writer];
      o << "  Write " << writer.second->to_string() << ",k=" << writer.first <<
	" uses name " << name << " " << v->to_string() << endl;
    }
  }
}

void Megatile::summarize_usib(ostream &o)
{
  o << "USIB summary: ";
  if (usib == NULL)
    o << "none" << endl;
  else {
    o << usib->size() << " locs (";
    bool first = true;
    foreach (u, llist<set<touch> *>, *usib) {
      if (first)
	first = false;
      else
	o << "; ";
      size_t reads = 0;
      count_if((*u)->begin(), (*u)->end(), touch_is_read, reads);
      size_t writes = (*u)->size() - reads;
      o << reads << "r/" << writes << "w";
    }
    o << ")" << endl;
  }
}

/* In the program as it stands now, c is a list of stmts to be done in order.
   Try to reorder them so that the first one is either before the megatile
   or after all of c. */
bool Megatile::try_reorder(llist<AC *> *&c, llist<OrderingConstraint *> *oc)
{
  /* to be done later */
  return false;
}

bool
Megatile::find_first_induced_in_offset_tile(Foreach *f, int j,
					    llist<OrderingConstraint *> *oc,
					    intvector *&v,
					    llist<int> *&new_m,
					    llist<Foreach *> *&new_loops)
{
  return induce_tile(f, oc, unit_vector(j, arity()), &v, new_m, new_loops);
}

bool Megatile::calculate_step(Foreach *f, int j,
			      llist<OrderingConstraint *> *oc,
			      llist<int> *&new_m, llist<Foreach *> *&new_loops)
{
  intvector *v = NULL;
  if (f->s == NULL)
    return false;
  if (!find_first_induced_in_offset_tile(f, j, oc, v, new_m, new_loops) ||
      v == NULL)
    return false;
  v = v->minus(f->s[0]);
  DBG(cout << "Induced step " << j << ": " << v->to_string() << endl);
  f->steps = ::cons(v, f->steps);
  // f->planes is not used in induced tilings.
  // f->planes = ::cons(new TilingPlane(v), f->planes);
  return true;
}

bool Megatile::calculate_steps(Foreach *f, llist<OrderingConstraint *> *oc,
			       llist<int> *&new_m,
			       llist<Foreach *> *&new_loops)
{
  int a = arity();
  assert(f->IS_arity == a);
  for (int i = a; --i >= 0; )
    if (!calculate_step(f, i, oc, new_m, new_loops))
      return false;

  DBG(cout << "Induced ");
  DBG(f->show_mapping());

  return true;
}

/* Pull candidates out into a list.  Go through the list (sorted
   lexicographically) and add each elt, if it is legal according to
   oc/been_done, to new_order, been_done, and f->lines.  been_done is what
   nodes have been done in f.  oc is list of ordering constraints from
   f to f.  new_order is in the format of a Megatile's m field, but
   reversed.  Returns false iff there is difficulty converting the set
   of candidates to a finite list. */
bool Megatile::add_candidates(Foreach *f, Relation candidates,
			      Relation &been_done,
			      llist<OrderingConstraint *> *oc,
			      llist<int> *&new_order)
{
  llist<intvector *> *l;
  if (!set_to_llist(candidates, l))
    return false;
  if (l == NULL)
    return true;
  intvector **c = l->to_array();
  int s = l->size();
  qsort_intvectors(c, s);
  for (int i = 0; i < s; i++)
    if (no_oc_violated(oc, been_done, c[i])) {
      f->lines = extend(f->lines, ::cons(c[i]));
      new_order = ::cons(f->k, ::cons(n, new_order));
      DBG(cout << "  Added <" << n << ", " << f->k << ">" << endl);
      f->k++;
      been_done = Union(been_done, *intvector_to_singleton_set(c[i]));
    }

  ::free_all(l);
  delete[] c;
  return true;
}

/* Analogous to add_candidates(), but stop after finding the first
   legal candidate, if there is one.  Fails (returns false) iff the
   set of candidates is infinite. */
bool Megatile::get_first_candidate(Relation candidates,
				   Relation &been_done,
				   llist<OrderingConstraint *> *oc,
				   intvector **first)
{
  llist<intvector *> *l;
  if (!set_to_llist(candidates, l))
    return false;
  if (l == NULL)
    return true;
  intvector **c = l->to_array();
  int s = l->size();
  qsort_intvectors(c, s);
  for (int i = 0; i < s; i++)
    if (no_oc_violated(oc, been_done, c[i])) {
      *first = c[i];
      break;
    }

  ::free_all(l);
  delete[] c;
  return true;
}

bool Megatile::induce_tile0(Foreach *f, llist<OrderingConstraint *> *oc,
			    llist<int> *&new_m, llist<Foreach *> *&new_loops)
{
  return induce_tile(f, oc, NULL, NULL, new_m, new_loops);
}

/* Operates in two modes.  If offset is NULL, then it induces a new
and improved Megatile by stepping through all the nodes in the current
tile0 of this Megatile, and interleaving nodes from f as they become
ready.  If offset is non-NULL, it operates similarly but with three big
differences: 1) the offset (in tile space) is applied, so we're
inducing a tile other than tile0, 2) we only find the first node
in the induced tile, and then we stop (after setting *first), and
3) we do not modify f.  Returns whether it succeeds. */
bool Megatile::induce_tile(Foreach *f, llist<OrderingConstraint *> *oc,
			   intvector *offset, intvector **first,
			   llist<int> *&new_m, llist<Foreach *> *&new_loops)
{
#define FAIL(x) do { failstr = (x); goto fail; } while (0)
#define SUCCEED() do { result = true; goto done; } while (0)
  string failstr;
  bool result = false;

  if (offset == NULL) {
    f->k = 0;
    f->lines = NULL;
    f->planes = NULL;
    f->steps = NULL;
    f->widths = NULL;
    f->unroll = 0; /* meaningless for induced tile */
    f->s = NULL;
    f->induced = true;
  } else {
    *first = NULL;
  }

  /* b[i] is what has been done in loop i (b[n] is what's been done in f) */
  Relation *b = new Relation [n + 1];

  /* ocx[i] is the list of constraints between loop i and f */
  llist<OrderingConstraint *> **ocx =
    new llist<OrderingConstraint *> * [n + 1];

  {
    Relation ready; /* what is ready in f */

    /* Suppose we've done tiles prior to tile0 in all loops in this. */
    for (int i = 0; i < n; i++) {
      b[i] = (*loops)[i]->tiles_prior_to_tile0();
      if (offset)
	b[i] = translate_set(b[i], (*loops)[i]->mapping(offset));
      ocx[i] = filter_OrderingConstraints(oc, (*loops)[i], f);
      Relation x = copy(compute_ready_set(b[i], ocx[i]));
      ready = (i == 0) ? x : Intersection(ready, x);
    }
    ocx[n] = filter_OrderingConstraints(oc, f, f);
    /* Assume we've done in f everything that we could have done. */
    b[n] = remove_nodes_that_violate_self_deps(ready, ocx[n]);
    printrel(ready, "Ready before stepping through megatile:");
    /* Note that at this point b[n] is not necessarily self-consistent.
       We may be assuming we've done a node x that requires some node y
       that is not in b[n]. */
  }

  llist<int> *new_order = NULL;
  for (iterator i = begin(); !i.isDone(); i.next()) {
    /* do a node */
    int l = i.loop_number();
    int v = i.v_number();
    if (l < n) {
      DBG(cout << "<" << l << ", " << v << ">" << endl);
      if (!offset)
	new_order = ::cons(v, ::cons(l, new_order));
      intvector *point_to_do = i.v();
      if (offset)
	point_to_do = point_to_do->plus((*loops)[l]->mapping(offset));
      b[l] = Union(b[l], *intvector_to_singleton_set(point_to_do));
      /* Recompute the ready set */
      Relation candidates = copy(compute_ready_set(b[l], ocx[l]));
      for (int j = 0; j < n + 1; j++)
	if (j != l)
	  candidates =
	    Intersection(candidates,
			 copy(compute_ready_set(b[j], ocx[j])));
      /* Candidates are what is now ready minus what was done before */
      candidates = Difference(candidates, copy(b[n]));
#if 0
      cout << "candidates after adding " << i.v()->to_string() << " of loop " <<
	l << ":";
      printrel(candidates, "");
#endif

      if (offset) {
	if (!get_first_candidate(candidates, b[n], ocx[n], first))
	  FAIL("candidates are infinite");
	if (*first)
	  SUCCEED();
      } else if (!add_candidates(f, candidates, b[n], ocx[n], new_order))
	FAIL("candidates are infinite");
    }
  }

  /* We get here if offset is NULL and we succeed in inducing tile0 */
  /* Modifying f a little is OK, but don't modify this because adding
     f to this still may not work. */
  f->s = f->lines->to_array();
  new_m = dreverse(new_order);
  new_loops = append(loops, ::cons(f));
  result = true;

  /* For the purposes of printing, temporarily pretend we've got f fully
     incorporated into this, which we don't yet. */
  DBG({
    n++;
    swap(m, new_m);
    swap(loops, new_loops);
    print(cout);
    swap(m, new_m);
    swap(loops, new_loops);
    n--;
  });

 done:
  delete[] b;
  delete[] ocx;
  return result;

 fail:
  DBG(cout << "Unable to induce tile: " << failstr << endl);
  result = false;
  goto done;
#undef FAIL
#undef SUCCEED
}

/* True iff parallelepiped at the origin contains p.  Defined so that
   there is no double counting of points on the faces, 
   i.e., we use 0 <= x < 1, for certain x's. 
   Normals to the faces must be provided by the caller, though they are
   just a function of l. */
static bool parallelepiped_contains(llist<intvector *> *l,
				    intvector **normal, 
				    intvector const *p)
{
  int j = 0;
  for (ListIterator<intvector *> i(l); !i.isDone(); i.next()) {
    intvector const *v = *i;
    intvector const *n = normal[j++];
    int x = n->dot(p);
    if (!(x >= 0 && x < n->dot(v)))
      return false;
  }
  return true;
}

/* If space were tiled with parallelepipeds whose edges are congruent
   to the elts of l, then how many elements of Z^n would be inside 
   each tile? */
static int size_of_parallelepiped(llist<intvector *> *l)
{
  size_t n = l->front()->size(); /* # of dimensions */
  assert(n == l->size());
  if (n == 1)
    return abs((*(l->front()))[0]);
  intvector **normal = new intvector * [n];
  for (int i = 0; i < (int) n; i++) {
    normal[i] = find_vector_perp_to_all(l, i);
    int d = normal[i]->dot((*l)[i]);
    if (d < 0)
      normal[i] = normal[i]->destructive_product(-1);
    else
      assert(d > 0);
  }

  /* Compute bounding box on parallelepiped starting at the origin */
  BoundingBox b;
  llist<llist<intvector *> *> *z = powerset(l); /* partial leak */
  while (z != NULL) {
    b.insert(sum_intvectors(z->front(), n));
    z = z->free();
  }
  /* Count number of points in the parallelepiped */
  int k = 0;
  for (BoxIter i = b.start(); !i.is_done(); i.next())
    if (parallelepiped_contains(l, normal, *i))
      k++;
  delete[] normal;
  return k;
}

bool Megatile::legal_everywhere(Foreach *f, llist<OrderingConstraint *> *oc,
				llist<int> *new_m, llist<Foreach *> *new_loops)
{
#define FAIL(x) do { failstr = (x); goto fail; } while (0)
#define VFAIL(x, v)							\
    do {								\
      if (verbose) {							\
	failstr = "see above";						\
	cout << "Problem: " << (x) << "---more detail:" << endl;	\
	{ v }								\
        goto fail;							\
      } else {								\
	FAIL(x);							\
      }									\
    } while (0)

  string failstr;
  bool result = true;
  bool verbose = true;

  /* Just for now, pretend f is part of this.  Undo these before returning. */
  swap(new_m, m);
  swap(new_loops, loops);
  n++;

  int i, arity = f->IS_arity;

  /* b[i] is what has been done in loop i (the last loop being f) */
  Relation *b = new Relation [n];

  /* ocx[i] is the list of constraints between loop i and f */
  llist<OrderingConstraint *> **ocx =
    new llist<OrderingConstraint *> * [n];

  /* We're going to reason about the generic tile (GT for short).  GT
     is the tile {a0, a1, ..., } in tile space, where the a's are
     symbolic variables ("free variables" in the Omega library). */
  Free_Var_Decl **a = new Free_Var_Decl * [arity];

  for (i = 0; i < arity; i++) {
    char s[10];
    sprintf(s, "v%d", i);
    a[i] = new Free_Var_Decl(s);
  }

  {
    Relation ready; /* what is ready in f */

    /* Suppose we've done tiles prior to GT in all loops in this. */
    for (i = 0; i < n - 1; i++) {
      b[i] = (*loops)[i]->tiles_prior_to_GT(a);
      ocx[i] = filter_OrderingConstraints(oc, (*loops)[i], f);
      Relation x = copy(compute_ready_set(b[i], ocx[i]));
      ready = (i == 0) ? x : Intersection(ready, x);
    }
    ocx[i] = filter_OrderingConstraints(oc, f, f);
    b[i] = f->tiles_prior_to_GT(a);
    if (!f->must_be_subset(copy(b[i]), copy(ready)))
      VFAIL("set of nodes done prior to GT is not a subset of the ready set",
	    show_subset_failure(b[i], ready););
    Relation self_ready = copy(compute_ready_set(b[i], ocx[i]));
    if (!f->must_be_subset(copy(b[i]), copy(self_ready)))
      VFAIL("set of nodes done prior to GT is not self-consistent",
	    show_subset_failure(b[i], self_ready););
  }

  for (iterator i = begin(); !i.isDone(); i.next()) {
    /* do a node */
    int l = i.loop_number();
    int v = i.v_number();
    DBG(cout << "<" << l << ", " << v << ">" << endl);
    Relation point_to_do = i.generic(a);
    /* Only when i is from f do we need to verify correctness. */
    if (i.loop() == f) {
      /* Fail if we've done it before or if it is not ready. */
      if (Intersection(copy(b[l]), copy(point_to_do)).is_satisfiable())
	FAIL("Attempting to do a point that is already done");
      Relation ready = copy(compute_ready_set(b[l], ocx[l]));
      for (int j = 0; j < n; j++)
	if (j != l)
	  ready =
	    Intersection(ready,
			 copy(compute_ready_set(b[j], ocx[j])));
      if (!Must_Be_Subset(copy(point_to_do), ready))
	FAIL("Attempting to do a point that is not ready");
    }
    b[l] = Union(b[l], point_to_do);
  }
  /* Success! */

 done:
  delete[] a;
  delete[] b;
  for (i = 0; i < n; i++)
    ::free_all(ocx[i]);
  delete[] ocx;

  swap(new_m, m);
  swap(new_loops, loops);
  n--;
  return result;

 fail:
  DBG(cout << "megatile failed verification: " << failstr << endl);
  result = false;
  goto done;
#undef FAIL
#undef VFAIL
}

/* Verify two things: that the tiling for f covers all of space, and that
   the order specified by the megatile is legal everywhere in space.
   If both pass, then the specified megatiling is legal.
   If f is unordered, we could probably get by with only testing the first.
 */
bool Megatile::verify(Foreach *f, llist<OrderingConstraint *> *oc,
		      llist<int> *new_m, llist<Foreach *> *new_loops)
{
#define FAIL(x) do { failstr = (x); goto fail; } while (0)
  string failstr;
  bool result = true;

  DBG(cout << endl << "Verifying legality of megatile" << endl); 

  /* Is f's tile the right size to cover all of space? */

  int s = size_of_parallelepiped(f->steps);
  // DBG(cout << "Size of tile0 is " << s << endl);
  if (s != f->k)
    FAIL("f->k is " + i2s(f->k) +
	 " but f's tile0 size is " + i2s(s));

  DBG(cout << "size of tile0 is OK" << endl);

  /* We now know f's tile is the right size, but it does not
     necessarily cover all of space, because, for example, there could
     be some points in space hit twice and others not at all.  We
     check for repeats in legal_everywhere(), and if none are found,
     then we know all of space is covered. */

  /* Is it legal everywhere, with no node done twice? */
  result = legal_everywhere(f, oc, new_m, new_loops);

  if (result)
    DBG(cout << "tile is legal everywhere" << endl << endl);

 done:
  return result;
 fail:
  DBG(cout << "megatile failed verification: " << failstr << endl);
  result = false;
  goto done;
#undef FAIL
}

bool Megatile::add(Foreach *f, llist<int> *new_m, llist<Foreach *> *new_loops)
{
  ::free_all(m);
  m = new_m;
  loops = new_loops;
  n++;
  return true;
}

/* Attempt to merge x into this.  Returns whether it succeeds.  If
   successful, destructively modifies this. */
bool Megatile::merge(AC *x, llist<OrderingConstraint *> *oc)
{
  if (x->is_Foreach() && !x->is_tiled_Foreach()) {
    Foreach *f = (Foreach *) x; 
    if (f->nesting_allows_tiling()) {
      DBG(cout << "Attempting to merge in loop at " << f->source_code_location <<
	endl);
      f->IS_arity = f->highest_arity_iteration_space();
      if (f->IS_arity == arity()) {
	llist<int> *new_m = NULL;
	llist<Foreach *> *new_loops = NULL;
	return induce_tile0(f, oc, new_m, new_loops) &&
	  calculate_steps(f, oc, new_m, new_loops) &&
	  verify(f, oc, new_m, new_loops) &&
	  add(f, new_m, new_loops);
      } else
	DBG(cout << "Can't merge in loop at "  << f->source_code_location <<
	    " because its arity is " << f->IS_arity << endl);
    } else {
      DBG(cout << "Nesting of loop at "  << f->source_code_location <<
	  " does not allow tiling" << endl);
    }
  }
  return false;
}

void Megatile::mark_loops_as_tiled() const
{
  /* mark all loops as tiled */
  llist<Foreach *> *l = loops;
  while (l != NULL) {
    l->front()->tiled = true;
    l = l->tail();
  }
}

void Megatile::remove_stmts(Block *b) const
{
  llist<Foreach *> *l = loops;
  while (l != NULL) {
    b->remove_stmt(l->front());
    l = l->tail();
  }
}

/* l is the tiled loop into which we try to merge subsequent stmts (c).
   If successful, return true and modify the block b to reflect the changes. */
static bool construct_megatile(Foreach *l, Block *b,
			       llist<AC *> *c, llist<OrderingConstraint *> *oc)
{
  bool success = false;
  Megatile *m = new Megatile(l);
  while (c != NULL) {
    if (m->merge(c->front(), oc)) {
      success = true;
      c = c->tail();
    }
    else if (c->tail() != NULL && m->try_reorder(c, oc)) ;
    else
      break;
  }
  if (success) {
    m->mark_loops_as_tiled();
    /* insert megatile into program; remove loops in megatile from program */
    int i = b->find_stmt(l);
    b->replace_stmt(i, m);
    m->remove_stmts(b);
  }
  return success;
}			       

///////////////////////////////////////////////////////////////////////////
// find_block_which_stmt_is_in(AC *s) returns NULL unless s is directly
//  a child of a Block
///////////////////////////////////////////////////////////////////////////
static Block *find_block_which_stmt_is_in(AC *s)
{
  if (s == NULL || s->parent == NULL)
    return NULL;
  return s->parent->find_block_which_stmt_is_in(s);
}

Block *AC::find_block_which_stmt_is_in(AC *s)
{
  return NULL;
}

Block *Block::find_block_which_stmt_is_in(AC *s)
{
  return this;
}
/////////////////////////////////////////////////////////////////////////////

static llist<Block *> *find_immediately_following_blocks(Block *b)
{
  Block *p = find_block_which_stmt_is_in(b);
  if (p == NULL)
    return NULL;
  llist<Block *> *result = NULL;
  llist<AC *> *x;
  for (x = p->statements(); x != NULL; x = x->tail())
    if (x->front() == b) break;
  assert(x != NULL);
  while ((x = x->tail()) != NULL) {
    if (x->front()->is_Block())
      result = cons((Block *) x->front(), result);
    else
      break;
  }
  return dreverse(result);
}

static llist<AC *> *find_subsequent_stmts(AC *s)
{
  DBG(cout << "find_subsequent_stmts() called on:" << endl);
  DBG(s->print(cout, 0));
  llist<AC *> *result = NULL;
  Block *b = find_block_which_stmt_is_in(s);
  DBG({
    cout << "enclosing block is:" << endl;
    if (b == NULL)
      cout << "NULL" << endl;
    else
      b->print(cout, 0);
  });
  if (b == NULL)
    return NULL;
  llist<Block *> *bb = find_immediately_following_blocks(b);
  result = copy(b->statements());
  while (bb != NULL) {
    result = extend(result, copy(bb->front()->statements()));
    bb = bb->tail();
  }
  while (result != NULL)
    if (result->front() == s) {
      DBG({
	cout << "subsequent stmts (" << result->size() - 1 << " of them):" <<
	  endl;
	for (llist<AC *> *i = result->tail(); i != NULL; i = i->tail())
	  i->front()->print(cout, 0);
      });
      return result->tail();
    } else
      result = result->tail();
  return result;
}

/* Return index of s in this, or -1 if not found. */
int Block::find_stmt(AC *s) const
{
  llist<AC *> *l = stmts;
  int i = 0;
  while (l != NULL)
    if (s == l->front())
      return i;
    else
      i++;
      
  return -1;
}

/////////////////////////////////////////////////////////////////////////////

/* Assumes there is at most one tiled (but not megatiled) loop and
   finds it.  Then creates a list of subsequent stmts that might form
   a megatile with the tiled loop.  Then tries to make a megatile.
   Returns true if a megatile is successfully created. */
bool Block::megatile(llist<OrderingConstraint *> *oc)
{
  for (llist<AC *> const *i = stmts; i != NULL; i = i->tail())
    if (i->front()->is_tiled_Foreach()) {
      llist<AC *> *c = find_subsequent_stmts(i->front());
      return construct_megatile((Foreach *) i->front(), this, c, oc);
    }

  return false;
}

bool AC::megatile(llist<OrderingConstraint *> *oc)
{
  return false;
}

/* l is loop number; k is step.  Step k in this megatile should be a
   node from loop l.  If step k is the first time that loop l is used
   in this then return 0.  If step k is the second time that loop l is
   used in this then return 1.  etc. */
int Megatile::which_node_from_loop(int l, int k)
{
  int count = 0;
  for (iterator i = begin(); k-- > 0; i.next())
    if (i.loop_number() == l)
      ++count;
  return count;
}

/* Return the number of nodes in loop number l that are used in this. */
/* Exception: if l is negative then return the total count of all loops. */
int Megatile::count(int l) const
{
  int count = 0;
  for (iterator i = begin(); !i.isDone(); i.next())
    if (i.loop_number() == l || l < 0)
      ++count;
  return count;
}

int Megatile::size() const { return count(-1); }

/* Sets f and v; caller should delete[] them. */
void Megatile::summarize(const Foreach **& f, const intvector **& v, int &s)
{
  s = size();
  f = (const Foreach **) new Foreach * [s];
  v = (const intvector **) new intvector * [s];
  int k = 0;
  for (iterator i = begin(); !i.isDone(); i.next(), k++) {
    f[k] = i.loop();
    v[k] = i.v();
  }
}

/* Return the bounding box of the tile for foreach l. */
BoundingBox Megatile::bb(int l) const 
{
  BoundingBox b;
  for (iterator i = begin(); !i.isDone(); i.next())
    if (i.loop_number() == l)
      b.insert(i.v());
  return b;
}

string Megatile::short_summary() const
{
  string s("tiling of " + i2s(n) + " loop");
  s += (n == 1) ? ":" : "s:";
  for (int i = 0; i < n; i++)
    s += ' ' + i2s(count(i)) + " nodes from loop " +
      (*loops)[i]->source_code_location;
  return s;
}

void Megatile::print(ostream &o, int depth = 0) const
{
  indent2(o, depth);
  o << "begin MEGATILE over " << n << " loops" << endl;
  for (int i = 0; i < n; i++) {
    indent2(o, depth + 1);
    int counti = count(i);
    if (counti != 1)
      o << count(i) << " nodes from loop " << i <<
	" (" << (*loops)[i]->source_code_location << ") bb: " <<
	bb(i).to_string() << endl;
    else
      o << "1 node from loop " << i <<
	" (" << (*loops)[i]->source_code_location << ")" << endl;
    indent2(o, depth + 1);
    o << " steps:";
    for (int j = 0; j < arity(); j++)
      o << ' ' << (*(*loops)[i]->steps)[j]->to_string();
    o << endl;
  }
  o << endl;
  for (iterator i = begin(); !i.isDone(); i.next()) {
    indent2(o, depth + 1);
    o << "From " << i.loop()->source_code_location <<
      " do " << i.v()->to_string() << endl;
  }
  indent2(o, depth);
  o << "end MEGATILE" << endl;
}

